# sage_setup: distribution = sagemath-modules
# sage.doctest: needs sage.graphs          (because all doctests use the catalogs simplicial_complexes, cubical_complexes)
"""
Homology and cohomology with a basis

This module provides homology and cohomology vector spaces suitable
for computing cup products and cohomology operations.

REFERENCES:

- [GDR2003]_
- [GDR1999]_

AUTHORS:

- John H. Palmieri, Travis Scrimshaw (2015-09)
"""
########################################################################
#       Copyright (C) 2015 John H. Palmieri <palmieri@math.washington.edu>
#                          Travis Scrimshaw <tscrimsh at umn.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#
#                  https://www.gnu.org/licenses/
########################################################################

from sage.algebras.steenrod.steenrod_algebra import SteenrodAlgebra
from sage.misc.cachefunc import cached_method
from sage.categories.algebras import Algebras
from sage.categories.category import Category
from sage.categories.left_modules import LeftModules
from sage.categories.right_modules import RightModules
from sage.categories.modules import Modules
from sage.combinat.free_module import CombinatorialFreeModule
from sage.matrix.constructor import matrix
from sage.modules.free_module_element import vector
from sage.sets.family import Family

try:
    from sage.topology.simplicial_complex import SimplicialComplex
    from sage.topology.simplicial_set import SimplicialSet_arbitrary
    from sage.topology.cubical_complex import CubicalComplex
except ImportError:
    SimplicialComplex = SimplicialSet_arbitrary = CubicalComplex = ()


class HomologyVectorSpaceWithBasis(CombinatorialFreeModule):
    r"""
    Homology (or cohomology) vector space.

    This provides enough structure to allow the computation of cup
    products and cohomology operations. See the class
    :class:`CohomologyRing` (which derives from this) for examples.

    It also requires field coefficients (hence the "VectorSpace" in
    the name of the class).

    .. NOTE::

        This is not intended to be created directly by the user, but
        instead via the methods
        :meth:`~sage.topology.cell_complex.GenericCellComplex.homology_with_basis` and
        :meth:`~sage.topology.cell_complex.GenericCellComplex.cohomology_ring`
        for the class of :class:`cell
        complexes<sage.topology.cell_complex.GenericCellComplex>`.

    INPUT:

    - ``base_ring`` -- must be a field
    - ``cell_complex`` -- the cell complex whose homology we are
      computing
    - ``cohomology`` -- boolean (default: ``False``); if ``True``, return
      the cohomology as a module
    - ``category`` -- (optional) a subcategory of modules with basis

    EXAMPLES:

    Homology classes are denoted by ``h_{d,i}`` where ``d`` is the
    degree of the homology class and ``i`` is their index in the list
    of basis elements in that degree. Cohomology classes are denoted
    ``h^{1,0}``::

        sage: RP2 = cubical_complexes.RealProjectivePlane()
        sage: RP2.homology_with_basis(GF(2))
        Homology module of Cubical complex with 21 vertices and 81 cubes
         over Finite Field of size 2
        sage: RP2.cohomology_ring(GF(2))
        Cohomology ring of Cubical complex with 21 vertices and 81 cubes
         over Finite Field of size 2
        sage: simplicial_complexes.Torus().homology_with_basis(QQ)
        Homology module of Minimal triangulation of the torus
         over Rational Field

    To access a basis element, use its degree and index (0 or 1 in the 1st
    cohomology group of a torus)::

        sage: H = simplicial_complexes.Torus().cohomology_ring(QQ)
        sage: H.basis(1)
        Finite family {(1, 0): h^{1,0}, (1, 1): h^{1,1}}
        sage: x = H.basis()[1,0]; x
        h^{1,0}
        sage: y = H.basis()[1,1]; y
        h^{1,1}
        sage: 2*x-3*y
        2*h^{1,0} - 3*h^{1,1}

    You can compute cup products of cohomology classes::

        sage: x.cup_product(y)
        -h^{2,0}
        sage: y.cup_product(x)
        h^{2,0}
        sage: x.cup_product(x)
        0

    This works with simplicial, cubical, and `\Delta`-complexes, and
    also simplicial sets::

        sage: Torus_c = cubical_complexes.Torus()
        sage: H = Torus_c.cohomology_ring(GF(2))
        sage: x,y = H.basis(1)
        sage: x.cup_product(x)
        0
        sage: x.cup_product(y)
        h^{2,0}
        sage: y.cup_product(y)
        0

        sage: Klein_d = delta_complexes.KleinBottle()
        sage: H = Klein_d.cohomology_ring(GF(2))
        sage: u,v = sorted(H.basis(1))
        sage: u.cup_product(u)
        h^{2,0}
        sage: u.cup_product(v)
        0
        sage: v.cup_product(v)
        h^{2,0}

    An isomorphism between the rings for the cubical model and the
    `\Delta`-complex model can be obtained by sending `x` to `u+v`,
    `y` to `v`. ::

        sage: # needs sage.groups
        sage: X = simplicial_sets.RealProjectiveSpace(6)
        sage: H_X = X.cohomology_ring(GF(2))
        sage: a = H_X.basis()[1,0]
        sage: a**6
        h^{6,0}
        sage: a**7
        0

    All products of positive-dimensional elements in a suspension
    should be zero::

        sage: # needs sage.groups
        sage: Y = X.suspension()
        sage: H_Y = Y.cohomology_ring(GF(2))
        sage: b = H_Y.basis()[2,0]
        sage: b**2
        0
        sage: B = sorted(H_Y.basis())[1:]
        sage: B
        [h^{2,0}, h^{3,0}, h^{4,0}, h^{5,0}, h^{6,0}, h^{7,0}]
        sage: import itertools
        sage: [a*b for (a,b) in itertools.combinations(B, 2)]
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

    The basis elements in the simplicial complex case have been chosen
    differently; apply the change of basis `x \mapsto a + b`, `y \mapsto
    b` to see the same product structure. ::

        sage: Klein_s = simplicial_complexes.KleinBottle()
        sage: H = Klein_s.cohomology_ring(GF(2))
        sage: a,b = H.basis(1)
        sage: a.cup_product(a)
        0
        sage: a.cup_product(b)
        h^{2,0}
        sage: (a+b).cup_product(a+b)
        h^{2,0}
        sage: b.cup_product(b)
        h^{2,0}
    """
    def __init__(self, base_ring, cell_complex, cohomology=False,
                 category=None) -> None:
        """
        Initialize ``self``.

        EXAMPLES::

            sage: RP2 = simplicial_complexes.ProjectivePlane()
            sage: H = RP2.homology_with_basis(QQ)
            sage: TestSuite(H).run()
            sage: H = RP2.homology_with_basis(GF(2))
            sage: TestSuite(H).run()
            sage: H = RP2.cohomology_ring(GF(2))
            sage: TestSuite(H).run()
            sage: H = RP2.cohomology_ring(GF(5))
            sage: TestSuite(H).run()
            sage: H = simplicial_complexes.ComplexProjectivePlane().cohomology_ring()
            sage: TestSuite(H).run()
        """
        # phi is the associated chain contraction.
        # M is the homology chain complex.
        phi, M = cell_complex.algebraic_topological_model(base_ring)
        if cohomology:
            phi = phi.dual()
            # We only need the rank of M in each degree, and since
            # we're working over a field, we don't need to dualize M
            # if working with cohomology.
        category = Modules(base_ring).WithBasis().Graded().FiniteDimensional().or_subcategory(category)
        self._contraction = phi
        self._complex = cell_complex
        self._cohomology = cohomology
        self._graded_indices = {deg: range(M.free_module_rank(deg))
                                for deg in range(cell_complex.dimension()+1)}
        indices = [(deg, i) for deg in self._graded_indices
                   for i in self._graded_indices[deg]]
        CombinatorialFreeModule.__init__(self, base_ring, indices, category=category)

    def basis(self, d=None):
        """
        Return (the degree ``d`` homogeneous component of) the basis
        of this graded vector space.

        INPUT:

        - ``d`` -- (optional) the degree

        EXAMPLES::

            sage: RP2 = simplicial_complexes.ProjectivePlane()
            sage: H = RP2.homology_with_basis(QQ)
            sage: H.basis()
            Finite family {(0, 0): h_{0,0}}
            sage: H.basis(0)
            Finite family {(0, 0): h_{0,0}}
            sage: H.basis(1)
            Finite family {}
            sage: H.basis(2)
            Finite family {}
        """
        if d is None:
            return Family(self._indices, self.monomial)
        else:
            indices = [(d, i) for i in self._graded_indices.get(d, [])]
            return Family(indices, self.monomial)

    def degree_on_basis(self, i):
        r"""
        Return the degree of the basis element indexed by ``i``.

        EXAMPLES::

            sage: H = simplicial_complexes.Torus().homology_with_basis(GF(7))
            sage: H.degree_on_basis((2,0))
            2
        """
        return i[0]

    def contraction(self):
        r"""
        The chain contraction associated to this homology computation.

        That is, to work with chain representatives of homology
        classes, we need the chain complex `C` associated to the cell
        complex, the chain complex `H` of its homology (with trivial
        differential), chain maps `\pi: C \to H` and `\iota: H \to C`,
        and a chain contraction `\phi` giving a chain homotopy between
        `1_C` and `\iota \circ \pi`.

        OUTPUT: `\phi`

        See :class:`~sage.homology.chain_homotopy.ChainContraction` for information
        about chain contractions, and see
        :func:`~sage.homology.algebraic_topological_model.algebraic_topological_model`
        for the construction of this particular chain contraction `\phi`.

        EXAMPLES::

            sage: H = simplicial_complexes.Simplex(2).homology_with_basis(QQ)
            sage: H.contraction()
            Chain homotopy between:
              Chain complex endomorphism of Chain complex with at most 3 nonzero terms over Rational Field
              and Chain complex endomorphism of Chain complex with at most 3 nonzero terms over Rational Field

        From the chain contraction, one can also recover the maps `\pi`
        and `\iota`::

            sage: phi = H.contraction()
            sage: phi.pi()
            Chain complex morphism:
              From: Chain complex with at most 3 nonzero terms over Rational Field
              To: Chain complex with at most 1 nonzero terms over Rational Field
            sage: phi.iota()
            Chain complex morphism:
              From: Chain complex with at most 1 nonzero terms over Rational Field
              To: Chain complex with at most 3 nonzero terms over Rational Field
        """
        return self._contraction

    def complex(self):
        """
        The cell complex whose homology is being computed.

        EXAMPLES::

            sage: H = simplicial_complexes.Simplex(2).homology_with_basis(QQ)
            sage: H.complex()
            The 2-simplex
        """
        return self._complex

    def _repr_(self) -> str:
        """
        EXAMPLES::

            sage: simplicial_complexes.Torus().homology_with_basis(QQ)
            Homology module of Minimal triangulation of the torus
             over Rational Field
        """
        if self._cohomology:
            base = "Cohomology"
        else:
            base = "Homology"
        return base + " module of {} over {}".format(self._complex, self.base_ring())

    def _repr_term(self, i) -> str:
        """
        Return ``'h_{i[0],i[1]}'`` for homology, ``'h^{i[0],i[1]}'`` for
        cohomology, for the basis element indexed by ``i``.

        EXAMPLES::

            sage: H = simplicial_complexes.Torus().homology_with_basis(QQ)
            sage: H.basis()[1,0] # indirect doctest
            h_{1,0}
            sage: latex(H.basis()[1,1]) # indirect doctest
            h_{1,1}
            sage: co = simplicial_complexes.KleinBottle().cohomology_ring(GF(2))
            sage: co.basis()[1,0] # indirect doctest
            h^{1,0}
        """
        sym = '^' if self._cohomology else '_'
        return 'h{}{{{},{}}}'.format(sym, i[0], i[1])

    _latex_term = _repr_term

    @cached_method
    def _to_cycle_on_basis(self, i):
        r"""
        Return the (co)cycle representative of the basis element
        indexed by ``i``.

        .. SEEALSO::

            :meth:`HomologyVectorSpaceWithBasis.Element.to_cocycle`

        EXAMPLES::

            sage: S2 = simplicial_complexes.Sphere(2)
            sage: H = S2.homology_with_basis(QQ)
            sage: H._to_cycle_on_basis((2,0))
            -(0, 1, 2) + (0, 1, 3) - (0, 2, 3) + (1, 2, 3)

            sage: S2.cohomology_ring(QQ)._to_cycle_on_basis((2,0))
            \chi_(1, 2, 3)
            sage: S2.cohomology_ring(QQ)._to_cycle_on_basis((0,0))
            \chi_(0,) + \chi_(1,) + \chi_(2,) + \chi_(3,)

            sage: RP3 = simplicial_complexes.RealProjectiveSpace(3)
            sage: H = RP3.cohomology_ring(GF(2))
            sage: H._to_cycle_on_basis((0,0))
            \chi_(1,) + \chi_(2,) + \chi_(3,) + \chi_(4,) + \chi_(5,) + \chi_(6,)
             + \chi_(7,) + \chi_(8,) + \chi_(9,) + \chi_(10,) + \chi_(11,)
            sage: H._to_cycle_on_basis((1,0))
            \chi_(2, 4) + \chi_(2, 5) + \chi_(2, 8) + \chi_(2, 10)
            + \chi_(3, 4) + \chi_(3, 6) + \chi_(3, 8) + \chi_(3, 9)
            + \chi_(4, 5) + \chi_(4, 6) + \chi_(4, 11) + \chi_(5, 7)
            + \chi_(5, 9) + \chi_(6, 7) + \chi_(6, 10) + \chi_(7, 8)
            + \chi_(9, 10)
            sage: H._to_cycle_on_basis((2,0))
            \chi_(3, 5, 9) + \chi_(3, 6, 10) + \chi_(3, 9, 10)
            + \chi_(4, 5, 7) + \chi_(4, 5, 9) + \chi_(4, 6, 7) + \chi_(6, 7, 10)
            sage: H._to_cycle_on_basis((3,0))
            \chi_(5, 6, 7, 8)
        """
        vec = self.contraction().iota().in_degree(i[0]).column(i[1])
        chains = self.complex().n_chains(i[0], self.base_ring(),
                                         cochains=self._cohomology)
        return chains.from_vector(vec)

    def dual(self):
        r"""
        Return the dual space.

        If ``self`` is homology, return the cohomology ring. If
        ``self`` is cohomology, return the homology as a vector space.

        EXAMPLES::

            sage: T = simplicial_complexes.Torus()
            sage: hom = T.homology_with_basis(GF(2))
            sage: coh = T.cohomology_ring(GF(2))
            sage: hom.dual() is coh
            True
            sage: coh.dual() is hom
            True
        """
        if is_GF2(self.base_ring()):
            if self._cohomology:
                return HomologyVectorSpaceWithBasis_mod2(self.base_ring(),
                                                         self.complex())
            return CohomologyRing_mod2(self.base_ring(), self.complex())
        if self._cohomology:
            return HomologyVectorSpaceWithBasis(self.base_ring(),
                                                self.complex(),
                                                not self._cohomology)
        return CohomologyRing(self.base_ring(), self.complex())

    def _test_duality(self, **options):
        r"""
        Test if the ordered bases for homology and cohomology are compatible.
        Return nothing if the test succeeds.

        This checks whether each evaluation map `H^n \otimes H_n \to
        k` is represented by the identity matrix, in terms of the
        chosen bases.

        TESTS::

            sage: T = simplicial_complexes.Torus()
            sage: K = T.suspension()
            sage: K.set_immutable()
            sage: H = K.cohomology_ring(QQ)
            sage: H._test_duality()

            sage: simplicial_complexes.RandomComplex(8, 2, .2).homology_with_basis(GF(2))._test_duality()
            sage: simplicial_complexes.RandomComplex(8, 2, .4).homology_with_basis(GF(2))._test_duality()
            sage: simplicial_complexes.RandomComplex(8, 2, .6).homology_with_basis(GF(2))._test_duality()

            sage: simplicial_complexes.RandomComplex(12, 3, .5).homology_with_basis(GF(2))._test_duality() # long time
        """
        tester = self._tester(**options)
        dual = self.dual()
        dims = [a[0] for a in self._indices]
        for dim in range(max(*dims, tester._max_runs) + 1):
            n = len(self.basis(dim))
            m = matrix(n, n, [a.eval(b) for a in self.basis(dim) for b in dual.basis(dim)])
            tester.assertEqual(m, 1, f"error in dimension {dim}")

    class Element(CombinatorialFreeModule.Element):
        def to_cycle(self):
            r"""
            (Co)cycle representative of this homogeneous (co)homology class.

            EXAMPLES::

                sage: S2 = simplicial_complexes.Sphere(2)
                sage: H = S2.homology_with_basis(QQ)
                sage: h20 = H.basis()[2,0]; h20
                h_{2,0}
                sage: h20.to_cycle()
                -(0, 1, 2) + (0, 1, 3) - (0, 2, 3) + (1, 2, 3)

            Chains are written as linear combinations of simplices
            `\sigma`. Cochains are written as linear combinations of
            characteristic functions `\chi_{\sigma}` for those
            simplices::

                sage: S2.cohomology_ring(QQ).basis()[2,0].to_cycle()
                \chi_(1, 2, 3)
                sage: S2.cohomology_ring(QQ).basis()[0,0].to_cycle()
                \chi_(0,) + \chi_(1,) + \chi_(2,) + \chi_(3,)
            """
            if not self.is_homogeneous():
                raise ValueError("only defined for homogeneous elements")
            return sum(c * self.parent()._to_cycle_on_basis(i) for i, c in self)

        def eval(self, other):
            r"""
            Evaluate ``self`` at ``other``.

            INPUT:

            - ``other`` -- an element of the dual space; if ``self``
              is an element of cohomology in dimension `n`, then
              ``other`` should be an element of homology in dimension
              `n`, and vice versa

            This just calls the :meth:`~sage.homology.chains.Cochains.Element.eval`
            method on the representing chains and cochains.

            EXAMPLES::

                sage: T = simplicial_complexes.Torus()
                sage: homology = T.homology_with_basis(QQ)
                sage: cohomology = T.cohomology_ring(QQ)
                sage: a1, a2 = homology.basis(1)
                sage: alpha1, alpha2 = cohomology.basis(1)
                sage: a1.to_cycle()
                (0, 3) - (0, 6) + (3, 6)
                sage: alpha1.to_cycle()
                -\chi_(1, 3) - \chi_(1, 4) - \chi_(2, 3) - \chi_(2, 4) - \chi_(2, 5) + \chi_(3, 6)
                sage: a1.eval(alpha1)
                1
                sage: alpha2.to_cycle()
                \chi_(1, 3) + \chi_(1, 4) + \chi_(1, 6) + \chi_(2, 4) - \chi_(4, 5) + \chi_(5, 6)
                sage: alpha2.eval(a1)
                0
                sage: (2 * alpha2).eval(a1 + a2)
                2
            """
            if not self or not other:
                return self.base_ring().zero()
            if self.parent()._cohomology:
                return self.to_cycle().eval(other.to_cycle())
            else:
                return other.to_cycle().eval(self.to_cycle())


class HomologyVectorSpaceWithBasis_mod2(HomologyVectorSpaceWithBasis):
    r"""
    Homology vector space mod 2.

    Based on :class:`HomologyVectorSpaceWithBasis`, with Steenrod
    operations included.

    .. NOTE::

        This is not intended to be created directly by the user, but
        instead via the method
        :meth:`~sage.topology.cell_complex.GenericCellComplex.homology_with_basis`
        for the class of :class:`cell
        complexes<sage.topology.cell_complex.GenericCellComplex>`.

    .. TODO::

        Implement Steenrod operations on (co)homology at odd primes,
        and thereby implement this class over `\GF{p}` for any `p`.

    INPUT:

    - ``base_ring`` -- must be the field ``GF(2)``
    - ``cell_complex`` -- the cell complex whose homology we are
      computing
    - ``category`` -- (optional) a subcategory of modules with basis

    This does not include the ``cohomology`` argument present for
    :class:`HomologyVectorSpaceWithBasis`: use
    :class:`CohomologyRing_mod2` for cohomology.

    EXAMPLES:

    Mod 2 cohomology operations are defined on both the left and the
    right::

        sage: # needs sage.groups
        sage: RP4 = simplicial_sets.RealProjectiveSpace(5)
        sage: H = RP4.homology_with_basis(GF(2))
        sage: x4 = H.basis()[4,0]
        sage: x4 * Sq(1)
        h_{3,0}
        sage: Sq(1) * x4
        h_{3,0}
        sage: Sq(2) * x4
        h_{2,0}
        sage: Sq(3) * x4
        h_{1,0}
        sage: Sq(0,1) * x4
        h_{1,0}
        sage: x4 * Sq(0,1)
        h_{1,0}
        sage: Sq(3) * x4
        h_{1,0}
        sage: x4 * Sq(3)
        0
    """
    def __init__(self, base_ring, cell_complex, category=None) -> None:
        """
        Initialize ``self``.

        EXAMPLES::

            sage: H = simplicial_complexes.Torus().homology_with_basis(GF(2))
            sage: TestSuite(H).run()
            sage: H = simplicial_complexes.Sphere(3).homology_with_basis(GF(2))
            sage: TestSuite(H).run()
        """
        if not is_GF2(base_ring):
            raise ValueError
        category = Modules(base_ring).WithBasis().Graded().FiniteDimensional().or_subcategory(category)
        category = Category.join((category,
                                  LeftModules(SteenrodAlgebra(2)),
                                  RightModules(SteenrodAlgebra(2))))
        HomologyVectorSpaceWithBasis.__init__(self, base_ring,
                                              cell_complex,
                                              cohomology=False,
                                              category=category)

    class Element(HomologyVectorSpaceWithBasis.Element):

        def _acted_upon_(self, a, self_on_left):
            r"""
            Define multiplication of ``self`` by ``a``, an
            element of the Steenrod algebra.

            INPUT:

            - ``a`` -- an element of the mod 2 Steenrod algebra
            - ``self_on_left`` -- ``True`` if we are computing ``self * a``,
              otherwise ``a * self``

            Algorithm: use the action of the Steenrod algebra `A` on
            cohomology to construct the action on homology. That is,
            given a right action of `A` on `H^*`,

            .. MATH::

                \phi_L: H^* \otimes A \to H^*

            we define (a la Boardman [Boa1982]_, p. 190)

            .. MATH::

                S'' \phi_L: A \otimes H_* \to H_*

            using the formula

            .. MATH::

                \langle (S'' \phi) (f \otimes a), x \rangle
                = \langle f, \phi_L (a \otimes x) \rangle,

            for `f \in H_m`, `a \in A^n`, and `x \in
            H^{m-n}`. Somewhat more succinctly, we define the action `f
            \cdot a` by

            .. MATH::

                (f \cdot a) (x) = f (a \cdot x).

            So given `f` (a.k.a. ``self``) and `a`, we compute `f (a
            \cdot x)` for all basis elements `x` in `H^{m-n}`,
            yielding a vector indexed by those basis elements. Since
            our basis for homology is dual to the basis for
            cohomology, we can then use the homology basis to convert
            the vector to an element of `H_{m-n}`.

            This gives a right module structure. To get a left module
            structure, use the right module structure after applying
            the antipode to `a`.

            EXAMPLES::

                sage: # needs sage.groups
                sage: RP5 = simplicial_sets.RealProjectiveSpace(5)
                sage: H = RP5.homology_with_basis(GF(2))
                sage: x5 = list(H.basis(5))[0]
                sage: Sq(1) * x5
                0
                sage: Sq(2) * x5
                h_{3,0}
                sage: x5 * Sq(2)
                h_{3,0}

            TESTS::

                sage: # needs sage.groups
                sage: RP4 = simplicial_sets.RealProjectiveSpace(5)
                sage: H = RP4.homology_with_basis(GF(2))
                sage: x4 = H.basis()[4,0]
                sage: (Sq(1) * Sq(2)) * x4 != 0
                True
                sage: (Sq(1) * Sq(2)) * x4 == Sq(1) * (Sq(2) * x4)
                True
                sage: x4 * (Sq(2) * Sq(1)) == (x4 * Sq(2)) * Sq(1)
                True
                sage: 1 * x4
                h_{4,0}
                sage: x4 * 0
                0
            """
            # Handle field elements first.
            ret = CombinatorialFreeModule.Element._acted_upon_(self, a, self_on_left)
            if ret is not None:  # did the scalar action
                return ret
            m = self.degree()
            n = a.degree()
            if m <= n:
                return self.parent().zero()

            if not self_on_left:  # i.e., module element on left
                a = a.antipode()
            P = self.parent()
            return P._from_dict({x.support()[0]: self.eval(a * x)
                                 for x in sorted(self.parent().dual().basis(m-n))})


class CohomologyRing(HomologyVectorSpaceWithBasis):
    """
    The cohomology ring.

    .. NOTE::

        This is not intended to be created directly by the user, but
        instead via the
        :meth:`cohomology ring<sage.topology.cell_complex.GenericCellComplex.cohomology_ring>`
        of a :class:`cell
        complex<sage.topology.cell_complex.GenericCellComplex>`.

    INPUT:

    - ``base_ring`` -- must be a field
    - ``cell_complex`` -- the cell complex whose homology we are
      computing
    - ``category`` -- (optional) a subcategory of modules with basis

    EXAMPLES::

        sage: CP2 = simplicial_complexes.ComplexProjectivePlane()
        sage: H = CP2.cohomology_ring(QQ)
        sage: H.basis(2)
        Finite family {(2, 0): h^{2,0}}
        sage: x = H.basis(2)[2,0]

    The product structure is the cup product::

        sage: x.cup_product(x)
        -h^{4,0}
        sage: x * x
        -h^{4,0}
    """
    def __init__(self, base_ring, cell_complex, category=None) -> None:
        """
        Initialize ``self``.

        EXAMPLES::

            sage: RP2 = simplicial_complexes.ProjectivePlane()
            sage: H = RP2.cohomology_ring(GF(5))
            sage: TestSuite(H).run()
            sage: T = simplicial_complexes.Torus()
            sage: H = T.cohomology_ring(QQ)
            sage: TestSuite(H).run()
        """
        if category is None:
            category = Algebras(base_ring).WithBasis().Graded().FiniteDimensional()
        HomologyVectorSpaceWithBasis.__init__(self, base_ring, cell_complex, True, category)

    def _repr_(self) -> str:
        """
        EXAMPLES::

            sage: simplicial_complexes.Torus().cohomology_ring(QQ)
            Cohomology ring of Minimal triangulation of the torus
             over Rational Field
        """
        return "Cohomology ring of {} over {}".format(self._complex, self.base_ring())

    @cached_method
    def one(self):
        """
        The multiplicative identity element.

        EXAMPLES::

            sage: H = simplicial_complexes.Torus().cohomology_ring(QQ)
            sage: H.one()
            h^{0,0}
            sage: all(H.one() * x == x == x * H.one() for x in H.basis())
            True
        """
        one = self.base_ring().one()
        d = {(0, i): one for i in self._graded_indices[0]}
        return self._from_dict(d, remove_zeros=False)

    @cached_method
    def product_on_basis(self, li, ri):
        r"""
        The cup product of the basis elements indexed by ``li`` and ``ri``
        in this cohomology ring.

        INPUT:

        - ``li``, ``ri`` -- index of a cohomology class

        .. SEEALSO::

            :meth:`CohomologyRing.Element.cup_product` -- the
            documentation for this method describes the algorithm.

        EXAMPLES::

            sage: RP3 = simplicial_complexes.RealProjectiveSpace(3)
            sage: H = RP3.cohomology_ring(GF(2))
            sage: c = H.basis()[1,0]
            sage: c.cup_product(c).cup_product(c) # indirect doctest
            h^{3,0}

            sage: T = simplicial_complexes.Torus()
            sage: x,y = T.cohomology_ring(QQ).basis(1)
            sage: x.cup_product(y)
            -h^{2,0}
            sage: x.cup_product(x)
            0

            sage: one = T.cohomology_ring(QQ).basis()[0,0]
            sage: x.cup_product(one)
            h^{1,0}
            sage: one.cup_product(y) == y
            True
            sage: one.cup_product(one)
            h^{0,0}
            sage: x.cup_product(y) + y.cup_product(x)
            0

        This also works with cubical complexes::

            sage: T = cubical_complexes.Torus()
            sage: x,y = T.cohomology_ring(QQ).basis(1)
            sage: x.cup_product(y)
            h^{2,0}
            sage: x.cup_product(x)
            0

        `\Delta`-complexes::

            sage: T_d = delta_complexes.Torus()
            sage: a,b = T_d.cohomology_ring(QQ).basis(1)
            sage: a.cup_product(b)
            h^{2,0}
            sage: b.cup_product(a)
            -h^{2,0}
            sage: RP2 = delta_complexes.RealProjectivePlane()
            sage: w = RP2.cohomology_ring(GF(2)).basis()[1,0]
            sage: w.cup_product(w)
            h^{2,0}

        and simplicial sets::

            sage: from sage.topology.simplicial_set_examples import RealProjectiveSpace
            sage: RP5 = RealProjectiveSpace(5)                                          # needs sage.groups
            sage: x = RP5.cohomology_ring(GF(2)).basis()[1,0]                           # needs sage.groups
            sage: x**4                                                                  # needs sage.groups
            h^{4,0}

        A non-connected example::

            sage: K = cubical_complexes.Torus().disjoint_union(cubical_complexes.Torus())
            sage: a,b,c,d = K.cohomology_ring(QQ).basis(1)
            sage: x,y = K.cohomology_ring(QQ).basis(0)
            sage: a.cup_product(x) == a
            True
            sage: a.cup_product(y)
            0
        """
        scomplex = self.complex()
        base_ring = self.base_ring()
        deg_left = li[0]
        deg_right = ri[0]
        deg_tot = deg_left + deg_right
        left_cycle = self._to_cycle_on_basis(li)
        right_cycle = self._to_cycle_on_basis(ri)
        n_chains_left = scomplex.n_chains(deg_left, base_ring)
        n_chains_right = scomplex.n_chains(deg_right, base_ring)

        result = {}
        H = scomplex.homology_with_basis(base_ring)
        for gamma_index in H._graded_indices.get(deg_tot, []):
            gamma_coeff = base_ring.zero()
            for cell, coeff in H._to_cycle_on_basis((deg_tot, gamma_index)):
                for (c, left_cell, right_cell) in scomplex.alexander_whitney(cell, deg_left):
                    if c:
                        left = n_chains_left(left_cell)
                        right = n_chains_right(right_cell)
                        gamma_coeff += c * coeff * left_cycle.eval(left) * right_cycle.eval(right)
            if gamma_coeff != base_ring.zero():
                result[(deg_tot, gamma_index)] = gamma_coeff
        return self._from_dict(result, remove_zeros=False)

    class Element(HomologyVectorSpaceWithBasis.Element):
        def cup_product(self, other):
            r"""
            Return the cup product of this element and ``other``.

            Algorithm: see González-Díaz and Réal [GDR2003]_, p. 88.
            Given two cohomology classes, lift them to cocycle
            representatives via the chain contraction for this
            complex, using
            :meth:`~HomologyVectorSpaceWithBasis.Element.to_cycle`. In
            the sum of their dimensions, look at all of the homology
            classes `\gamma`: lift each of those to a cycle
            representative, apply the Alexander-Whitney diagonal map
            to each cell in the cycle, evaluate the two cocycles on
            these factors, and multiply. The result is the value of
            the cup product cocycle on this homology class. After this
            has been done for all homology classes, since homology and
            cohomology are dual, one can tell which cohomology class
            corresponds to the cup product.

            .. SEEALSO::

                :meth:`CohomologyRing.product_on_basis`

            EXAMPLES::

                sage: RP3 = simplicial_complexes.RealProjectiveSpace(3)
                sage: H = RP3.cohomology_ring(GF(2))
                sage: c = H.basis()[1,0]
                sage: c.cup_product(c)
                h^{2,0}
                sage: c * c * c
                h^{3,0}

            We can also take powers::

                sage: RP2 = simplicial_complexes.RealProjectivePlane()
                sage: a = RP2.cohomology_ring(GF(2)).basis()[1,0]
                sage: a**0
                h^{0,0}
                sage: a**1
                h^{1,0}
                sage: a**2
                h^{2,0}
                sage: a**3
                0

            A non-connected example::

                sage: K = cubical_complexes.Torus().disjoint_union(cubical_complexes.Sphere(2))
                sage: a,b = K.cohomology_ring(QQ).basis(2)
                sage: a**0
                h^{0,0} + h^{0,1}
            """
            return self * other


class CohomologyRing_mod2(CohomologyRing):
    r"""
    The mod 2 cohomology ring.

    Based on :class:`CohomologyRing`, with Steenrod operations included.

    .. NOTE::

        This is not intended to be created directly by the user, but
        instead via the
        :meth:`cohomology ring<sage.topology.cell_complex.GenericCellComplex.cohomology_ring>`
        of a :class:`cell
        complex<sage.topology.cell_complex.GenericCellComplex>`.

    .. TODO::

        Implement Steenrod operations on (co)homology at odd primes,
        and thereby implement this class over `\GF{p}` for any `p`.

    INPUT:

    - ``base_ring`` -- must be the field ``GF(2)``
    - ``cell_complex`` -- the cell complex whose homology we are
      computing

    EXAMPLES:

    Mod 2 cohomology operations are defined on both the left and the
    right::

        sage: CP2 = simplicial_complexes.ComplexProjectivePlane()
        sage: Hmod2 = CP2.cohomology_ring(GF(2))
        sage: y = Hmod2.basis(2)[2,0]
        sage: y.Sq(2)
        h^{4,0}

        sage: # needs sage.groups
        sage: Y = simplicial_sets.RealProjectiveSpace(6).suspension()
        sage: H_Y = Y.cohomology_ring(GF(2))
        sage: b = H_Y.basis()[2,0]
        sage: b.Sq(1)
        h^{3,0}
        sage: b.Sq(2)
        0
        sage: c = H_Y.basis()[4,0]
        sage: c.Sq(1)
        h^{5,0}
        sage: c.Sq(2)
        h^{6,0}
        sage: c.Sq(3)
        h^{7,0}
        sage: c.Sq(4)
        0

    Cohomology can be viewed as a left module over the Steenrod
    algebra, and also as a right module::

        sage: # needs sage.groups
        sage: RP4 = simplicial_sets.RealProjectiveSpace(4)
        sage: H = RP4.cohomology_ring(GF(2))
        sage: x = H.basis()[1,0]
        sage: Sq(0,1) * x
        h^{4,0}
        sage: Sq(3) * x
        0
        sage: x * Sq(3)
        h^{4,0}
    """
    def __init__(self, base_ring, cell_complex) -> None:
        """
        Initialize ``self``.

        EXAMPLES::

            sage: RP2 = simplicial_complexes.ProjectivePlane()
            sage: H = RP2.cohomology_ring(GF(2))
            sage: TestSuite(H).run()
        """
        if not is_GF2(base_ring):
            raise ValueError("the base ring must be GF(2)")
        category = Algebras(base_ring).WithBasis().Graded().FiniteDimensional()
        category = Category.join((category,
                                  LeftModules(SteenrodAlgebra(2)),
                                  RightModules(SteenrodAlgebra(2))))
        CohomologyRing.__init__(self, base_ring, cell_complex, category=category)

    class Element(CohomologyRing.Element):
        def Sq(self, i):
            r"""
            Return the result of applying `Sq^i` to this element.

            INPUT:

            - ``i`` -- nonnegative integer

            .. WARNING::

                The main implementation is only for simplicial
                complexes and simplicial sets; cubical complexes are
                converted to simplicial complexes first. Note that
                this converted complex may be large and so computations
                may be slow. There is no implementation for
                `\Delta`-complexes.

            This cohomology operation is only defined in
            characteristic 2. Odd primary Steenrod operations are not
            implemented.

            Algorithm: see González-Díaz and Réal [GDR1999]_,
            Corollary 3.2.

            EXAMPLES::

                sage: RP2 = simplicial_complexes.RealProjectiveSpace(2)
                sage: x = RP2.cohomology_ring(GF(2)).basis()[1,0]
                sage: x.Sq(1)
                h^{2,0}

                sage: K = RP2.suspension()
                sage: K.set_immutable()
                sage: y = K.cohomology_ring(GF(2)).basis()[2,0]
                sage: y.Sq(1)
                h^{3,0}

                sage: # long time
                sage: # needs sage.groups
                sage: RP4 = simplicial_complexes.RealProjectiveSpace(4)
                sage: H = RP4.cohomology_ring(GF(2))
                sage: x = H.basis()[1,0]
                sage: y = H.basis()[2,0]
                sage: z = H.basis()[3,0]
                sage: x.Sq(1) == y
                True
                sage: z.Sq(1)
                h^{4,0}

            This calculation is much faster with simplicial sets (on
            one machine, 20 seconds with a simplicial complex, 4 ms
            with a simplicial set). ::

                sage: RP4_ss = simplicial_sets.RealProjectiveSpace(4)                   # needs sage.groups
                sage: z_ss = RP4_ss.cohomology_ring(GF(2)).basis()[3,0]                 # needs sage.groups
                sage: z_ss.Sq(1)                                                        # needs sage.groups
                h^{4,0}

            TESTS::

                sage: RP_cubical = cubical_complexes.RealProjectivePlane()
                sage: x = RP_cubical.cohomology_ring(GF(2)).basis()[1,0]
                sage: x.Sq(1)
                h^{2,0}
                sage: T = delta_complexes.Torus()
                sage: x = T.cohomology_ring(GF(2)).basis()[1,0]
                sage: x.Sq(1)
                Traceback (most recent call last):
                ...
                NotImplementedError: Steenrod squares are not implemented for this type of cell complex
            """
            P = self.parent()
            scomplex = P.complex()
            if isinstance(scomplex, CubicalComplex):
                # Convert cubical complex to simplicial complex, and
                # convert self to basis element in the new complex's
                # cohomology ring.
                scomplex = SimplicialComplex(scomplex, is_mutable=False)
                P = scomplex.cohomology_ring(self.base_ring())
                self = P.sum_of_terms(self.monomial_coefficients().items())
            if not isinstance(scomplex, (SimplicialComplex, SimplicialSet_arbitrary)):
                print(scomplex, isinstance(scomplex, SimplicialComplex))
                raise NotImplementedError('Steenrod squares are not implemented for '
                                          'this type of cell complex')
            scomplex = P.complex()
            base_ring = P.base_ring()
            if not is_GF2(base_ring):
                # This should never happen: the class should only be
                # instantiated in characteristic 2.
                raise ValueError('Steenrod squares are only defined in characteristic 2')
            # We keep the same notation as in [GDR1999].
            # The trivial cases:
            if i == 0:
                # Sq^0 is the identity.
                return self

            # Construct each graded component of ``self``
            ret = P.zero()
            H = scomplex.homology_with_basis(base_ring)
            deg_comp = {}
            for index, coeff in self:
                d = deg_comp.get(index[0], {})
                d[index] = coeff
                deg_comp[index[0]] = d

            # Do the square on each graded component of ``self``.
            for j in deg_comp:
                # Make it into an actual element
                m = j + i
                if not P._graded_indices.get(m, []) or i > j:
                    continue
                elt = P._from_dict(deg_comp[j], remove_zeros=False)
                if i == j:
                    ret += elt.cup_product(elt)
                    continue

                n = j - i
                # Now assemble the indices over which the sums take place.
                # S(n) is defined to be floor((m+1)/2) + floor(n/2).
                S_n = (m+1) // 2 + n // 2
                if n == 0:
                    sums = [[S_n]]
                else:
                    sums = [[i_n] + l for i_n in range(S_n, m+1)
                            for l in sum_indices(n-1, i_n, S_n)]
                # At this point, 'sums' is a list of lists of the form
                # [i_n, i_{n-1}, ..., i_0]. (It is reversed from the
                # obvious order because this is closer to the order in
                # which the face maps will be applied.)  Now we sum over
                # these, according to the formula in [GDR1999], Corollary 3.2.
                result = {}
                cycle = elt.to_cycle()
                n_chains = scomplex.n_chains(j, base_ring)
                for gamma_index in H._graded_indices.get(m, []):
                    gamma_coeff = base_ring.zero()
                    for cell, coeff in H._to_cycle_on_basis((m, gamma_index)):
                        for indices in sums:
                            indices = list(indices)
                            left = cell
                            right = cell
                            # Since we are working with a simplicial complex, 'cell' is a simplex.
                            if not m % 2:
                                left_endpoint = m
                                while indices:
                                    right_endpoint = indices[0] - 1
                                    for k in range(left_endpoint, indices.pop(0), -1):
                                        left = scomplex.face(left, k)
                                    try:
                                        left_endpoint = indices[0] - 1
                                        for k in range(right_endpoint, indices.pop(0), -1):
                                            right = scomplex.face(right, k)
                                    except IndexError:
                                        pass
                                for k in range(right_endpoint, -1, -1):
                                    right = scomplex.face(right, k)
                            else:
                                right_endpoint = m
                                while indices:
                                    left_endpoint = indices[0] - 1
                                    try:
                                        for k in range(right_endpoint, indices.pop(0), -1):
                                            right = scomplex.face(right, k)
                                        right_endpoint = indices[0] - 1
                                    except IndexError:
                                        pass
                                    for k in range(left_endpoint, indices.pop(0), -1):
                                        left = scomplex.face(left, k)
                                for k in range(right_endpoint, -1, -1):
                                    right = scomplex.face(right, k)

                            if ((hasattr(left, 'is_nondegenerate')
                                 and left.is_nondegenerate()
                                 and right.is_nondegenerate())
                                    or not hasattr(left, 'is_nondegenerate')):
                                left = n_chains(left)
                                right = n_chains(right)
                                gamma_coeff += coeff * cycle.eval(left) * cycle.eval(right)
                    if gamma_coeff != base_ring.zero():
                        result[(m, gamma_index)] = gamma_coeff
                ret += P._from_dict(result, remove_zeros=False)
            return ret

        def _acted_upon_(self, a, self_on_left):
            r"""
            Define multiplication of ``self`` by ``a``, an
            element of the Steenrod algebra.

            INPUT:

            - ``a`` -- an element of the mod 2 Steenrod algebra
            - ``self_on_left`` -- ``True`` if we are computing ``self * a``,
              otherwise ``a * self``

            Algorithm: for left multiplication by ``a``, since we have
            :meth:`Sq` to compute multiplication by a single generator
            `Sq^i`, first convert ``a`` to the Serre-Cartan basis ---
            that is, sums of products of the elements `Sq^i` --- and
            then apply :meth:`Sq` repeatedly. Right multiplication by
            ``a`` is the same as left multiplication by the antipode
            applied to ``a``.

            EXAMPLES::

                sage: # needs sage.groups
                sage: SRP4 = simplicial_sets.RealProjectiveSpace(4).suspension()
                sage: H = SRP4.cohomology_ring(GF(2))
                sage: x = H.basis()[2,0]
                sage: Sq(0,1) * x
                h^{5,0}
                sage: Sq(3) * x
                0
                sage: x * Sq(3)
                h^{5,0}

            TESTS::

                sage: # needs sage.groups
                sage: (Sq(2) * Sq(1)) * x == Sq(2) * (Sq(1) * x)
                True
                sage: x * (Sq(1) * Sq(2)) == (x * Sq(1)) * Sq(2)
                True
                sage: x * 1
                h^{2,0}
                sage: 0 * x
                0
            """
            # Handle field elements first.
            ret = CombinatorialFreeModule.Element._acted_upon_(self, a, self_on_left)
            if ret is not None:  # did the scalar action
                return ret
            if self_on_left:  # i.e., module element on left
                a = a.antipode()
            b = a.change_basis('adem')
            ans = self.parent().zero()
            mono_dict = b.monomial_coefficients()
            for seq in mono_dict:
                x = self
                for i in reversed(seq):
                    x = x.Sq(i)
                ans += mono_dict[seq] * x
            return ans

    def steenrod_module_map(self, deg_domain, deg_codomain, side='left'):
        r"""
        Return a component of the module structure map `A \otimes
        H \to H`, where `H` is this cohomology ring and `A` is the
        Steenrod algebra.

        INPUT:

        - ``deg_domain`` -- the degree of the domain in the cohomology
          ring

        - ``deg_codomain`` -- the degree of the codomain in the
          cohomology ring

        - ``side`` -- (default: ``'left'``) whether we are computing
          the action as a left module action or a right module

        We will write this with respect to the left action;
        for the right action, just switch all of the tensors.
        Writing `m` for ``deg_domain`` and `n` for ``deg_codomain``, this
        returns `A^{n-m} \otimes H^{m} \to H^{n}`, one single
        component of the map making `H` into an `A`-module.

        .. WARNING::

           This is only implemented in characteristic two. The main
           implementation is only for simplicial complexes and simplicial
           sets; cubical complexes are converted to simplicial complexes
           first. Note that this converted complex may be large and so
           computations may be slow. There is no implementation for
           `\Delta`-complexes.

        ALGORITHM:

        Use the Milnor basis for the truncated Steenrod
        algebra `A`, and for cohomology, use the basis with which it
        is equipped. For each pair of basis elements `a` and `h`,
        compute the product `a \otimes h`, and use this to assemble a
        matrix defining the action map via multiplication on the
        appropriate side. That is, if ``side`` is ``'left'``, return a
        matrix suitable for multiplication on the left, etc.

        EXAMPLES::

            sage: # needs sage.groups
            sage: RP4 = simplicial_sets.RealProjectiveSpace(4)
            sage: H = RP4.cohomology_ring(GF(2))
            sage: H.steenrod_module_map(1, 2)
            [1]
            sage: H.steenrod_module_map(1, 3)
            [0]
            sage: H.steenrod_module_map(1, 4, 'left')
            [1 0]
            sage: H.steenrod_module_map(1, 4, 'right')
            [1]
            [1]

        Products of projective spaces::

            sage: RP3 = simplicial_sets.RealProjectiveSpace(3)
            sage: K = RP3.product(RP3)
            sage: H = K.cohomology_ring(GF(2))
            sage: H
            Cohomology ring of RP^3 x RP^3 over Finite Field of size 2

        There is one column for each element `a \otimes b`, where `a`
        is a basis element for the Steenrod algebra and `b` is a basis
        element for the cohomology algebra.  There is one row for each
        basis element of the cohomology algebra. Unfortunately, the
        chosen basis for this truncated polynomial algebra is not the
        monomial basis::

            sage: x1, x2 = H.basis(1)
            sage: x1 * x1
            h^{2,0} + h^{2,1}
            sage: x2 * x2
            h^{2,2}
            sage: x1 * x2
            h^{2,0}

            sage: # needs sage.combinat
            sage: H.steenrod_module_map(1, 2)
            [1 0]
            [1 0]
            [0 1]
            sage: H.steenrod_module_map(1, 3, 'left')
            [0 0]
            [0 0]
            [0 0]
            [0 0]
            sage: H.steenrod_module_map(1, 3, 'right')
            [0 0 0 0]
            [0 0 0 0]
            sage: H.steenrod_module_map(2, 3)
            [0 0 0]
            [1 1 0]
            [0 0 0]
            [0 0 0]
        """
        side = side.lower()
        if side not in ['right', 'left']:
            raise ValueError('side must be either "left" or "right"')
        base_ring = self.base_ring()
        char = base_ring.characteristic()
        # Use sorted(...) so that there are reproducible orders on the bases.
        A_basis = sorted(SteenrodAlgebra(char).basis(deg_codomain - deg_domain))
        H_basis_dom = sorted(self.basis(deg_domain))
        H_basis_cod = self.basis(deg_codomain).keys()
        H_basis_cod = {elt: idx for idx, elt in enumerate(sorted(H_basis_cod))}
        entries = []
        for a in A_basis:
            for h in H_basis_dom:
                vec = vector(base_ring, len(H_basis_cod))
                if side == 'left':
                    x = a * h
                else:
                    x = h * a
                monos = x.monomial_coefficients()
                for seq in monos:
                    vec[H_basis_cod[seq]] = monos[seq]
                entries.extend(vec)
        # We built the matrix column by column, so now we take the
        # transpose.
        if side == 'left':
            return matrix(base_ring, len(A_basis) * len(H_basis_dom),
                          len(H_basis_cod), entries).transpose()
        return matrix(base_ring, len(A_basis) * len(H_basis_dom),
                          len(H_basis_cod), entries)


def sum_indices(k, i_k_plus_one, S_k_plus_one):
    r"""
    This is a recursive function for computing the indices for the
    nested sums in González-Díaz and Réal [GDR1999]_, Corollary 3.2.

    In the paper, given indices `i_n`, `i_{n-1}`, ..., `i_{k+1}`,
    given `k`, and given `S(k+1)`, the number `S(k)` is defined to be

    .. MATH::

        S(k) = -S(k+1) + floor(k/2) + floor((k+1)/2) + i_{k+1},

    and `i_k` ranges from `S(k)` to `i_{k+1}-1`. There are two special
    cases: if `k=0`, then `i_0 = S(0)`. Also, the initial case of
    `S(k)` is `S(n)`, which is set in the method :meth:`Sq` before
    calling this function. For this function, given `k`, `i_{k+1}`,
    and `S(k+1)`, return a list consisting of the allowable possible
    indices `[i_k, i_{k-1}, ..., i_1, i_0]` given by the above
    formula.

    INPUT:

    - ``k`` -- nonnegative integer
    - ``i_k_plus_one`` -- the positive integer `i_{k+1}`
    - ``S_k_plus_one`` -- the integer `S(k+1)`

    EXAMPLES::

        sage: from sage.homology.homology_vector_space_with_basis import sum_indices
        sage: sum_indices(1, 3, 3)
        [[1, 0], [2, 1]]
        sage: sum_indices(0, 4, 2)
        [[2]]
    """
    S_k = -S_k_plus_one + k//2 + (k+1)//2 + i_k_plus_one
    if k == 0:
        return [[S_k]]
    return [[i_k] + l for i_k in range(S_k, i_k_plus_one)
            for l in sum_indices(k-1, i_k, S_k)]


def is_GF2(R) -> bool:
    r"""
    Return ``True`` iff ``R`` is isomorphic to the field `\GF{2}`.

    EXAMPLES::

        sage: from sage.homology.homology_vector_space_with_basis import is_GF2
        sage: is_GF2(GF(2))
        True
        sage: is_GF2(GF(2, implementation='ntl'))                                       # needs sage.libs.ntl
        True
        sage: is_GF2(GF(3))
        False
    """
    return 2 == R.characteristic() == R.cardinality()



from sage.config import get_include_dirs


def test_cython_metaclass_header_found():
    dirs = get_include_dirs()
    assert any(
        (dir / "sage" / "cpython" / "cython_metaclass.h").is_file() for dir in dirs
    )


def test_get_include_dirs_returns_existing_dirs():
    dirs = get_include_dirs()
    for dir in dirs:
        assert dir.is_dir(), f"Directory {dir} does not exist"



# sage_setup: distribution = sagemath-modules
r"""
Quotients of free modules

AUTHORS:

- William Stein (2009): initial version

- Kwankyu Lee (2022-05): added quotient module over domain
"""

# ****************************************************************************
#       Copyright (C) 2009 William Stein <wstein@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.modules.free_module import (
    FreeModule_ambient,
    FreeModule_ambient_field,
    Module_free_ambient,
)

###############################################################################
#
# Quotients of ambient free modules over a domain
#
###############################################################################


class QuotientModule_free_ambient(Module_free_ambient):
    """
    Quotients of ambient free modules by a submodule.

    INPUT:

    - ``module`` -- an ambient free module

    - ``sub`` -- a submodule of ``module``

    EXAMPLES::

        sage: S.<x,y,z> = PolynomialRing(QQ)
        sage: M = S**2
        sage: N = M.submodule([vector([x - y, z]), vector([y*z, x*z])])
        sage: M.quotient_module(N)
        Quotient module by Submodule of Ambient free module of rank 2 over
         the integral domain Multivariate Polynomial Ring in x, y, z over Rational Field
         Generated by the rows of the matrix:
         [x - y     z]
         [  y*z   x*z]
    """
    def __init__(self, module, sub):
        """
        Create this quotient module of ``module`` by a submodule ``sub``.

        TESTS::

            sage: S.<x,y,z> = PolynomialRing(QQ)
            sage: M = S**2
            sage: N = M.submodule([vector([x - y, z]), vector([y*z, x*z])])
            sage: Q = M.quotient_module(N)
            sage: TestSuite(Q).run(skip=['_test_elements', '_test_pickling', '_test_zero'])
        """
        base_ring = module.base_ring()
        degree = module.degree()
        sparse = module.is_sparse()
        # We store these in order to retain the information of how this was constructed
        self._module = module
        self._sub = sub
        self.__hash = hash((self._module, self._sub))

        # We then convert the data into maps from free modules
        if isinstance(module, QuotientModule_free_ambient):
            self._free_cover = module.cover()
            C = self._free_cover.element_class
            v = [C(self._free_cover, x.list(), coerce=False, copy=False) for x in sub.gens()]
            w = [C(self._free_cover, x.list(), coerce=False, copy=False) for x in module.free_relations().gens()]
            self._relations = self._free_cover.submodule(v + w, check=False)
        else:  # Otherwise module should be a free module
            self._free_cover = module
            self._relations = sub

        Module_free_ambient.__init__(self, base_ring, degree=degree, sparse=sparse)

    def _repr_(self):
        r"""
        Return the string representation of this quotient module.

        EXAMPLES::

            sage: S.<x,y,z> = PolynomialRing(QQ)
            sage: M = S**2
            sage: N = M.submodule([vector([x - y, z]), vector([y*z , x*z])])
            sage: M.quotient_module(N)
            Quotient module by Submodule of Ambient free module of rank 2 over
             the integral domain Multivariate Polynomial Ring in x, y, z over Rational Field
             Generated by the rows of the matrix:
             [x - y     z]
             [  y*z   x*z]
        """
        return "Quotient module by %s" % self._sub

    def __hash__(self):
        r"""
        Return the hash of this quotient module.

        TESTS::

            sage: S.<x,y,z> = PolynomialRing(QQ)
            sage: M = S**2
            sage: N = M.submodule([vector([x - y, z]), vector([y*z, x*z])])
            sage: Q = M.quotient_module(N)
            sage: hash(Q) == hash((M,N))
            True
        """
        return self.__hash

    def gens(self) -> tuple:
        """
        Return the generators of this module.

        EXAMPLES::

            sage: S.<x,y,z> = PolynomialRing(QQ)
            sage: M = S**2
            sage: N = M.submodule([vector([x - y, z]), vector([y*z, x*z])])
            sage: Q = M.quotient_module(N)
            sage: Q.gens()
            ((1, 0), (0, 1))
        """
        return tuple(self(list(g)) for g in self._module.gens())

    def gen(self, i=0):
        """
        Return the `i`-th generator of this module.

        EXAMPLES::

            sage: S.<x,y,z> = PolynomialRing(QQ)
            sage: M = S**2
            sage: N = M.submodule([vector([x - y, z]), vector([y*z, x*z])])
            sage: Q = M.quotient_module(N)
            sage: Q.gen(0)
            (1, 0)
        """
        if i < 0 or i >= self._module.degree():
            raise ValueError('generator %s not defined' % i)
        return self.gens()[i]

    def _coerce_map_from_(self, M):
        """
        Return a coercion map from `M` to ``self``, or ``None``.

        TESTS::

            sage: S.<x,y,z> = PolynomialRing(QQ)
            sage: M = S**2
            sage: N = M.submodule([vector([x - y, z]), vector([y*z, x*z])])
            sage: Q = M.quotient_module(N)
            sage: Q.coerce_map_from(M)
            Coercion map:
              From: Ambient free module of rank 2 over the integral domain
                    Multivariate Polynomial Ring in x, y, z over Rational Field
              To:   Quotient module by Submodule of
                    Ambient free module of rank 2 over the integral domain
                    Multivariate Polynomial Ring in x, y, z over Rational Field
                    Generated by the rows of the matrix:
                    [x - y     z]
                    [  y*z   x*z]
        """
        if isinstance(M, FreeModule_ambient):
            return (self.base_ring().has_coerce_map_from(M.base_ring()) and
                    self.degree() == M.degree())
        from sage.modules.submodule import Submodule_free_ambient
        if isinstance(M, Submodule_free_ambient):
            return self._module.has_coerce_map_from(self.ambient_module())
        if (isinstance(M, QuotientModule_free_ambient)
                and M.free_cover() == self.free_cover()):
            try:
                return M.free_relations().is_submodule(self.free_relations())
            except NotImplementedError:
                pass

    def ambient_module(self):
        """
        Return ``self``, since ``self`` is ambient.

        EXAMPLES::

            sage: S.<x,y,z> = PolynomialRing(QQ)
            sage: M = S**2
            sage: N = M.submodule([vector([x - y, z]), vector([y*z, x*z])])
            sage: Q = M.quotient_module(N)
            sage: Q.ambient_module() is Q
            True
        """
        return self

    def cover(self):
        r"""
        Given this quotient space `Q = V/W`, return `V`.

        EXAMPLES::

            sage: S.<x,y,z> = PolynomialRing(QQ)
            sage: M = S**2
            sage: N = M.submodule([vector([x - y, z]), vector([y*z, x*z])])
            sage: Q = M.quotient_module(N)
            sage: Q.cover() is M
            True
        """
        return self._module

    V = cover

    def relations(self):
        r"""
        Given this quotient space `Q = V/W`, return `W`.

        EXAMPLES::

            sage: S.<x,y,z> = PolynomialRing(QQ)
            sage: M = S**2
            sage: N = M.submodule([vector([x - y, z]), vector([y*z, x*z])])
            sage: Q = M.quotient_module(N)
            sage: Q.relations() is N
            True
        """
        return self._sub

    W = relations

    def free_cover(self):
        r"""
        Given this quotient space `Q = V/W`, return the free module
        that covers `V`.

        EXAMPLES::

            sage: S.<x,y,z> = PolynomialRing(QQ)
            sage: M = S**2
            sage: N = M.submodule([vector([x - y, z]), vector([y*z, x*z])])
            sage: Q = M / N
            sage: NQ = Q.submodule([Q([1,x])])
            sage: QNQ = Q / NQ
            sage: QNQ.free_cover() is Q.free_cover() is M
            True

        Note that this is different than the immediate cover::

            sage: QNQ.cover() is Q
            True
            sage: QNQ.cover() is QNQ.free_cover()
            False
        """
        return self._free_cover

    V = cover

    def free_relations(self):
        r"""
        Given this quotient space `Q = V/W`, return the submodule that
        generates all relations of `Q`.

        When `V` is a free module, then this returns `W`. Otherwise this
        returns the union of `W` lifted to the cover of `V` and the relations
        of `V` (repeated until `W` is a submodule of a free module).

        EXAMPLES::

            sage: S.<x,y,z> = PolynomialRing(QQ)
            sage: M = S**2
            sage: N = M.submodule([vector([x - y, z]), vector([y*z, x*z])])
            sage: Q = M / N
            sage: NQ = Q.submodule([Q([1, x])])
            sage: QNQ = Q / NQ
            sage: QNQ.free_relations()
            Submodule of Ambient free module of rank 2 over the integral domain
            Multivariate Polynomial Ring in x, y, z over Rational Field
            Generated by the rows of the matrix:
            [    1     x]
            [x - y     z]
            [  y*z   x*z]

        Note that this is different than the defining relations::

            sage: QNQ.relations() is NQ
            True
            sage: QNQ.relations() == QNQ.free_relations()
            False
        """
        return self._relations


###############################################################################
#
# Quotients of ambient free modules over a field
#
###############################################################################

class FreeModule_ambient_field_quotient(FreeModule_ambient_field):
    """
    A quotient `V/W` of two vector spaces as a vector space.

    To obtain `V` or `W` use ``self.V()`` and ``self.W()``.

    EXAMPLES::

        sage: # needs sage.rings.number_field
        sage: k.<i> = QuadraticField(-1)
        sage: A = k^3; V = A.span([[1,0,i], [2,i,0]])
        sage: W = A.span([[3,i,i]])
        sage: U = V/W; U
        Vector space quotient V/W of dimension 1 over Number Field in i
         with defining polynomial x^2 + 1 with i = 1*I where
         V: Vector space of degree 3 and dimension 2 over Number Field in i
            with defining polynomial x^2 + 1 with i = 1*I
            Basis matrix:
            [ 1  0  i]
            [ 0  1 -2]
         W: Vector space of degree 3 and dimension 1 over Number Field in i
            with defining polynomial x^2 + 1 with i = 1*I
            Basis matrix:
            [    1 1/3*i 1/3*i]
        sage: U.V()
        Vector space of degree 3 and dimension 2 over Number Field in i
         with defining polynomial x^2 + 1 with i = 1*I
         Basis matrix:
         [ 1  0  i]
         [ 0  1 -2]
        sage: U.W()
        Vector space of degree 3 and dimension 1 over Number Field in i
         with defining polynomial x^2 + 1 with i = 1*I
         Basis matrix:
         [    1 1/3*i 1/3*i]
        sage: U.quotient_map()
        Vector space morphism represented by the matrix:
        [  1]
        [3*i]
        Domain:   Vector space of degree 3 and dimension 2 over Number Field in i
                  with defining polynomial x^2 + 1 with i = 1*I
                  Basis matrix:
                  [ 1  0  i]
                  [ 0  1 -2]
        Codomain: Vector space quotient V/W of dimension 1 over Number Field in i
                  with defining polynomial x^2 + 1 with i = 1*I where
                  V: Vector space of degree 3 and dimension 2 over Number Field in i
                     with defining polynomial x^2 + 1 with i = 1*I
                     Basis matrix:
                     [ 1  0  i]
                     [ 0  1 -2]
                  W: Vector space of degree 3 and dimension 1 over Number Field in i
                     with defining polynomial x^2 + 1 with i = 1*I
                     Basis matrix:
                     [    1 1/3*i 1/3*i]
        sage: Z = V.quotient(W)
        sage: Z == U
        True

    We create three quotient spaces and compare them::

        sage: A = QQ^2
        sage: V = A.span_of_basis([[1,0], [1,1]])
        sage: W0 = V.span([V.1, V.0])
        sage: W1 = V.span([V.1])
        sage: W2 = V.span([V.1])
        sage: Q0 = V/W0
        sage: Q1 = V/W1
        sage: Q2 = V/W2

        sage: Q0 == Q1
        False
        sage: Q1 == Q2
        True

    TESTS::

        sage: A = QQ^0; V = A.span([])  # corner case
        sage: W = A.span([])
        sage: U = V/W

        sage: loads(dumps(U)) == U
        True
        sage: type(loads(dumps(U)) )
        <class 'sage.modules.quotient_module.FreeModule_ambient_field_quotient_with_category'>
    """
    def __init__(self, domain, sub, quotient_matrix, lift_matrix, inner_product_matrix=None):
        """
        Create this quotient space, from the given domain, submodule,
        and quotient_matrix.

        EXAMPLES::

            sage: A = QQ^5; V = A.span_of_basis([[1,0,-1,1,1], [1,-1,0,2/3,3/4]]); V
            Vector space of degree 5 and dimension 2 over Rational Field
            User basis matrix:
            [  1   0  -1   1   1]
            [  1  -1   0 2/3 3/4]
            sage: W = V.span_of_basis([V.0 - 2/3*V.1]); W
            Vector space of degree 5 and dimension 1 over Rational Field
            User basis matrix:
            [1/3 2/3  -1 5/9 1/2]

        This creates a quotient vector space::

            sage: Q = V / W

        Behold the type of Q::

            sage: type(Q)
            <class 'sage.modules.quotient_module.FreeModule_ambient_field_quotient_with_category'>

        We do some consistency checks on the extra quotient and
        lifting structure of Q::

            sage: Q(V.0)
            (1)
            sage: Q(V.0 - 2/3*V.1)
            (0)
            sage: v = Q.lift(Q.0); v
            (1, 0, -1, 1, 1)
            sage: Q( v )
            (1)
        """
        base_field = domain.base_field()
        dimension = quotient_matrix.ncols()
        sparse = domain.is_sparse()
        self._sub = sub
        self._domain = domain
        self.__hash = hash((domain, sub))
        FreeModule_ambient_field.__init__(self, base_field, dimension, sparse)
        self.__quo_map = domain.Hom(self)(quotient_matrix)
        self.__quo_map.register_as_coercion()
        self.__lift_map = self.Hom(domain)(lift_matrix)

    def _repr_(self):
        r"""
        Return the rather verbose string representation of this quotient space V/W.

        EXAMPLES:

        We create a quotient vector space over a finite field::

            sage: # needs sage.rings.finite_rings
            sage: k.<a> = GF(9); A = k^3; V = A.span_of_basis([[1,0,a], [a,a,1]]); W = V.span([V.1])
            sage: Q = V/W

        Note the type::

            sage: type(Q)                                                               # needs sage.rings.finite_rings
            <class 'sage.modules.quotient_module.FreeModule_ambient_field_quotient_with_category'>

        The string representation mentions that this is a quotient
        `V/W`, that the quotient has dimension 1 and is over a finite
        field, and also describes `V` and `W`::

            sage: Q._repr_()                                                            # needs sage.rings.finite_rings
            'Vector space quotient V/W of dimension 1 over Finite Field in a of size 3^2 where\nV: Vector space of degree 3 and dimension 2 over Finite Field in a of size 3^2\nUser basis matrix:\n[1 0 a]\n[a a 1]\nW: Vector space of degree 3 and dimension 1 over Finite Field in a of size 3^2\nBasis matrix:\n[    1     1 a + 2]'
        """
        return "%s space quotient V/W of dimension %s over %s where\nV: %s\nW: %s" % (
            "Sparse vector" if self.is_sparse() else "Vector",
            self.dimension(), self.base_ring(),
            self.V(), self.W())

    def __hash__(self):
        """
        Return hash of this quotient space `V/W`, which is, by definition,
        the hash of the tuple `(V, W)`.

        EXAMPLES:

        We compute the hash of a certain 0-dimension quotient vector
        space::

            sage: A = QQ^2; V = A.span_of_basis([[1,0], [1,1]]); W = V.span([V.1, V.0])
            sage: Q = V/W; Q.dimension()
            0
            sage: hash(Q) == hash((V,W))
            True
        """
        return self.__hash

    def _element_constructor_(self, x):
        """
        Convert an element into this quotient space `V/W` if there is
        a way to make sense of it.

        An element converts into ``self`` if it can be converted into `V`,
        or if not at least if it can be made sense of as a list of
        length the dimension of ``self``.

        EXAMPLES:

        We create a 2-dimensional quotient of a 3-dimension ambient
        vector space::

            sage: M = QQ^3 / [[1,2,3]]

        A list of length 3 converts into ``QQ^3``, so it converts into
        `M`::

            sage: M([1,2,4])  #indirect doctest
            (-1/3, -2/3)
            sage: M([1,2,3])
            (0, 0)

        A list of length 2 converts into M, where here it just gives
        the corresponding linear combination of the basis for `M`::

            sage: M([1,2])
            (1, 2)
            sage: M.0 + 2*M.1
            (1, 2)

        Of course, elements of ``QQ^3`` convert into the quotient
        module as well. Here is a different example::

            sage: V = QQ^3; W = V.span([[1,0,0]]); Q = V/W
            sage: Q(V.0)
            (0, 0)
            sage: Q(V.1)
            (1, 0)
            sage: Q.0
            (1, 0)
            sage: Q.0 + V.1
            (2, 0)

        Here we start with something that is over ZZ, so it
        canonically coerces into ``QQ^3``, hence into ``self``::

            sage: Q((ZZ^3)([1,2,3]))
            (2, 3)

        TESTS:

        Ensure that :issue:`39507` is fixed::

            sage: V = VectorSpace(GF(2),3)
            sage: A = V.subspace([(0,1,0), (1,1,1)])
            sage: B = A.subspace([])
            sage: list(A/B)
            [(0, 0), (1, 0), (0, 1), (1, 1)]
        """
        if isinstance(x, self.element_class) and x.parent() is self:
            return x
        if isinstance(x, (list, tuple)) and len(x) == self._domain.rank():
            return self.__quo_map(self._domain.linear_combination_of_basis(x))
        return FreeModule_ambient_field._element_constructor_(self, x)

    def _coerce_map_from_(self, M):
        """
        Return a coercion map from `M` to ``self``, or None.

        EXAMPLES::

            sage: V = QQ^2 / [[1, 2]]
            sage: V.coerce_map_from(ZZ^2)
            Composite map:
              From: Ambient free module of rank 2 over the principal ideal domain Integer Ring
              To:   Vector space quotient V/W of dimension 1 over Rational Field where
                    V: Vector space of dimension 2 over Rational Field
                    W: Vector space of degree 2 and dimension 1 over Rational Field
                       Basis matrix:
                       [1 2]
              Defn:   Coercion map:
                      From: Ambient free module of rank 2 over the principal ideal domain Integer Ring
                      To:   Vector space of dimension 2 over Rational Field
                    then
                      Vector space morphism represented by the matrix:
                      [   1]
                      [-1/2]
                      Domain:   Vector space of dimension 2 over Rational Field
                      Codomain: Vector space quotient V/W of dimension 1 over Rational Field where
                                V: Vector space of dimension 2 over Rational Field
                                W: Vector space of degree 2 and dimension 1 over Rational Field
                                   Basis matrix:
                                   [1 2]

        Make sure :issue:`10513` is fixed (no coercion from an abstract
        vector space to an isomorphic quotient vector space)::

            sage: V = QQ^3 / [[1,2,3]]
            sage: V.coerce_map_from(QQ^2)
        """
        from sage.modules.free_module import FreeModule_ambient
        if (isinstance(M, FreeModule_ambient)
            and not (isinstance(M, FreeModule_ambient_field_quotient)
                     and self._sub == M._sub)):
            # No map between different quotients.
            # No map from quotient to abstract module.
            return None
        f = super(FreeModule_ambient_field, self)._coerce_map_from_(M)
        if f is not None:
            return f
        f = self._domain.coerce_map_from(M)
        if f is not None:
            return self.__quo_map * f
        return None

    def quotient_map(self):
        """
        Given this quotient space `Q = V / W`, return the natural quotient
        map from `V` to `Q`.

        EXAMPLES::

            sage: M = QQ^3 / [[1,2,3]]
            sage: M.quotient_map()
            Vector space morphism represented by the matrix:
            [   1    0]
            [   0    1]
            [-1/3 -2/3]
            Domain:   Vector space of dimension 3 over Rational Field
            Codomain: Vector space quotient V/W of dimension 2 over Rational Field where
                      V: Vector space of dimension 3 over Rational Field
                      W: Vector space of degree 3 and dimension 1 over Rational Field
                      Basis matrix:
                      [1 2 3]

            sage: M.quotient_map()( (QQ^3)([1,2,3]) )
            (0, 0)
        """
        return self.__quo_map

    def lift_map(self):
        r"""
        Given this quotient space `Q = V / W`, return a fixed choice of
        linear homomorphism (a section) from `Q` to `V`.

        EXAMPLES::

            sage: M = QQ^3 / [[1,2,3]]
            sage: M.lift_map()
            Vector space morphism represented by the matrix:
            [1 0 0]
            [0 1 0]
            Domain:   Vector space quotient V/W of dimension 2 over Rational Field where
                      V: Vector space of dimension 3 over Rational Field
                      W: Vector space of degree 3 and dimension 1 over Rational Field
                         Basis matrix:
                         [1 2 3]
            Codomain: Vector space of dimension 3 over Rational Field
        """
        return self.__lift_map

    def lift(self, x):
        r"""
        Lift element of this quotient `V / W` to `V` by applying
        the fixed lift homomorphism.

        The lift is a fixed homomorphism.

        EXAMPLES::

            sage: M = QQ^3 / [[1,2,3]]
            sage: M.lift(M.0)
            (1, 0, 0)
            sage: M.lift(M.1)
            (0, 1, 0)
            sage: M.lift(M.0 - 2*M.1)
            (1, -2, 0)
        """
        return self.__lift_map(x)

    def cover(self):
        r"""
        Given this quotient space `Q = V/W`, return `V`.

        EXAMPLES::

            sage: M = QQ^10 / [list(range(10)), list(range(2,12))]
            sage: M.cover()
            Vector space of dimension 10 over Rational Field
        """
        return self._domain

    V = cover

    def relations(self):
        r"""
        Given this quotient space `Q = V/W`, return `W`.

        EXAMPLES::

            sage: M = QQ^10 / [list(range(10)), list(range(2,12))]
            sage: M.relations()
            Vector space of degree 10 and dimension 2 over Rational Field
            Basis matrix:
            [ 1  0 -1 -2 -3 -4 -5 -6 -7 -8]
            [ 0  1  2  3  4  5  6  7  8  9]
        """
        return self._sub

    W = relations



# sage_setup: distribution = sagemath-schemes
"""
Utility routines for hypergeometric motives

These are functions that benefit significantly from Cythonization.
"""
from cpython cimport array
from cysignals.signals cimport sig_check
from sage.rings.integer cimport Integer

cpdef hgm_coeffs(long long p, unsigned int f,
                 int prec, gamma, m, int D,
                 gtable, int gtable_prec, bint use_longs):
    r"""
    Compute coefficients for the hypergeometric trace formula.

    This function is not intended for direct user access.

    TESTS::

        sage: from sage.modular.hypergeometric_motive import HypergeometricData as Hyp
        sage: from sage.modular.hypergeometric_misc import hgm_coeffs
        sage: H = Hyp(cyclotomic=([3],[4]))
        sage: H.euler_factor(2, 7, cache_p=True)                                        # needs sage.symbolic
        7*T^2 - 3*T + 1
        sage: gamma = H.gamma_array()
        sage: prec, gtable = H.gauss_table(7, 1, 2)
        sage: D = 1
        sage: hgm_coeffs(7, 1, 2, gamma, [0]*6, D, gtable, prec, False)
        [7, 2*7, 6*7, 7, 6, 4*7]

    Check issue from :issue:`28404`::

        sage: H = Hyp(cyclotomic=[[10,2],[1,1,1,1,1]])
        sage: u = H.euler_factor(2,79)  # indirect doctest                              # needs sage.symbolic
        sage: u.reverse().is_weil_polynomial()                                          # needs sage.symbolic
        True
    """
    from sage.rings.padics.factory import Zp

    cdef int gl, j, k, v, gv
    cdef long long i, l, q1, w, w1, w2, q2, r, r1
    cdef bint flip, use_longlongs

    q1 = p ** f - 1
    gl = len(gamma)
    cdef array.array gamma_array1 = array.array('i', gamma.keys())
    cdef array.array gamma_array2 = array.array('i', gamma.values())
    r_array = [0] * gl
    cdef array.array digit_count = array.array('i', [0]) * q1
    cdef array.array gtab2

    try:
        R = gtable[0].parent()
    except AttributeError:
        R = Zp(p, prec, "fixed-mod")
    # In certain cases, the reciprocals of the Gauss sums are reported
    # for efficiency.
    flip = (f == 1 and prec == 1 and gtable_prec == 1)
    ans = []
    Rz = R.zero()
    if use_longs:
        q2 = p ** prec
        try:
            gtab2 = gtable
        except TypeError:
            gtab2 = array.array('l', [0]) * q1
            for r in range(q1):
                gtab2[r] = gtable[r].lift() % q2
    else:
        use_longlongs = (Integer(p) ** prec < 2 ** 63)
        if use_longlongs:
            gtab2 = array.array('q', [0]) * q1
            try:
                for r in range(q1):
                    gtab2[r] = gtable[r]
            except TypeError:
                for r in range(q1):
                    gtab2[r] = gtable[r].lift()
    if f == 1:
        for r in range(q1):
            digit_count[r] = r
    else:
        for r in range(q1):
            r1 = r
            w = 0
            for i in range(f):
                w += r1 % p
                r1 //= p
            digit_count[r] = w
    ans = [None] * q1
    for r in range(q1):
        sig_check()
        # Skip this coefficient if we already have it by symmetry.
        if ans[r] is not None:
            continue
        # Determine whether this term is forced to be zero
        # for divisibility reasons. If so, skip the p-adic arithmetic.
        i = 0
        for k in range(gl):
            v = gamma_array1[k]
            gv = gamma_array2[k]
            r1 = v * r % q1
            r_array[k] = r1
            i += digit_count[r1] * gv
        i //= (p - 1)
        l = i + f * (D + m[0] - m[r])
        if l >= prec:
            ans[r] = Rz
        else:
            # Keep numerator and denominator separate for efficiency.
            if use_longs:
                w = 1
                w1 = 1
            else:
                u = R.one()
                u1 = R.one()
            for k in range(gl):
                gv = gamma_array2[k]
                r1 = r_array[k]
                if flip:
                    gv = -gv
                if use_longs:
                    w2 = gtab2[r1]  # cast to long long to avoid overflow
                    if gv > 0:
                        for j in range(gv):
                            w = w * w2 % q2
                    else:
                        for j in range(-gv):
                            w1 = w1 * w2 % q2
                elif use_longlongs:
                    w2 = gtab2[r1]
                    if gv > 0:
                        for j in range(gv):
                            u *= w2
                    else:
                        for j in range(-gv):
                            u1 *= w2
                else:
                    if gv > 0:
                        for j in range(gv):
                            u *= gtable[r1]
                    else:
                        for j in range(-gv):
                            u1 *= gtable[r1]
            if use_longs:
                u = R(w)
                u1 = R(w1)
            if i % 2:
                u = -u
            ans[r] = (u / u1) << l
        if f > 1:
            r1 = r
            for j in range(f-1):
                r1 = r1 * p % q1
                if ans[r1] is not None:
                    break
                ans[r1] = ans[r]
    if f == 1:
        return ans
    # Consolidate down to p-1 terms.
    ans2 = [Rz] * (p-1)
    for r1 in range(p-1):
        for r in range(r1, q1, p-1):
            ans2[r1] += ans[r]
    return ans2



# sage_setup: distribution = sagemath-modules
# sage.doctest: needs sage.modules
r"""
Finite dimensional graded commutative algebras

AUTHORS:

- Michael Jung (2021): initial version
"""
# ****************************************************************************
#       Copyright (C) 2021 Michael Jung <m.jung at vu.nl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from __future__ import annotations

from sage.categories.algebras import Algebras
from sage.combinat.free_module import CombinatorialFreeModule
from sage.combinat.integer_vector_weighted import WeightedIntegerVectors
from sage.misc.cachefunc import cached_method
from sage.misc.functional import is_even
from sage.rings.integer_ring import ZZ
from sage.sets.condition_set import ConditionSet
from sage.sets.disjoint_union_enumerated_sets import DisjointUnionEnumeratedSets


class FiniteGCAlgebra(CombinatorialFreeModule):
    r"""
    Finite dimensional graded commutative algebras.

    A finite dimensional graded commutative algebra `A` is an integer-graded
    algebra satisfying the super-algebra relation w.r.t. the degree modulo 2.
    More precisely, `A` has a graded ring structure

    .. MATH::

        A = \bigoplus_{i=0}^n A_i,

    where `n \in \NN` is the finite maximal degree, and the multiplication
    satisfies

    .. MATH::

        A_i \cdot A_j \subset \begin{cases}A_{i+j} & \text{if $i+j\leq n$}, \\
                                             0 & \text{if $i+j > n$},\end{cases}

    as well as the super-algebra relation

    .. MATH::

        x y = (-1)^{ij} y x

    for all homogeneous elements `x \in A_i` and `y \in A_j`.

    Such an algebra is multiplicatively generated by a set of single monomials
    `\{ x_1, \ldots, x_k \}`, where each `x_i` is given a certain degree
    `\mathrm{deg}(x_i)`. To that end, this algebra can be given a vector
    space basis, and the basis vectors are of the form `x_1^{w_1} \cdots x_n^{
    w_k}`, where `\sum_{i=1}^k \mathrm{deg}(x_i) \, w_i \leq n` and

    .. MATH::

        w_i \in \begin{cases} \ZZ_2 & \text{if $\mathrm{deg}(x_i)$ is odd}, \\
        \NN & \text{if $\mathrm{deg}(x_i)$ is even}. \end{cases}

    Typical examples of finite dimensional graded commutative algebras are
    cohomology rings over finite dimensional CW-complexes.

    INPUT:

    - ``base`` -- the base field
    - ``names`` -- (optional) names of the generators: a list of
      strings or a single string with the names separated by
      commas. If not specified, the generators are named "x0", "x1",...
    - ``degrees`` -- (optional) a tuple or list specifying the degrees
      of the generators; if omitted, each generator is given degree
      1, and if both ``names`` and ``degrees`` are omitted, an error is
      raised
    - ``max_degree`` -- the maximal degree of the graded algebra
    - ``mul_symbol`` -- (optional) symbol used for multiplication; if omitted,
      the string "*" is used
    - ``mul_latex_symbol`` -- (optional) latex symbol used for multiplication;
      if omitted, the empty string is used

    EXAMPLES::

        sage: A.<x,y,z,t> = GradedCommutativeAlgebra(QQ, degrees=(1,2,2,3), max_degree=6)
        sage: A
        Graded commutative algebra with generators ('x', 'y', 'z', 't')
         in degrees (1, 2, 2, 3) with maximal degree 6
        sage: t*x + x*t
        0
        sage: x^2
        0
        sage: x*t^2
        0
        sage: x*y^2 + z*t
        x*y^2 + z*t

    The generators can be returned with :meth:`algebra_generators`::

        sage: F = A.algebra_generators(); F
        Family (x, y, z, t)
        sage: [g.degree() for g in F]
        [1, 2, 2, 3]

    We can also return the basis::

        sage: list(A.basis())                                                           # needs sage.combinat
        [1, x, z, y, t, x*z, x*y, x*t, z^2, y*z, y^2, z*t, y*t, x*z^2, x*y*z, x*y^2]

    Depending on the context, the multiplication can be given a different
    symbol::

        sage: A.<x,y,z,t> = GradedCommutativeAlgebra(QQ, degrees=(1,2,6,6), max_degree=10,
        ....:                                        mul_symbol='⌣',
        ....:                                        mul_latex_symbol=r'\smile')
        sage: x*y^2 + x*t
        x⌣y^2 + x⌣t
        sage: latex(x*y^2 - z*x)
        x\smile y^{2} - x\smile z

    .. NOTE::

        Notice, when the argument ``max_degree`` in the global namespace is
        omitted, an instance of the class
        :class:`sage.algebras.commutative_dga.GCAlgebra` is created instead::

            sage: # needs sage.combinat sage.libs.singular
            sage: A.<x,y,z,t> = GradedCommutativeAlgebra(QQ, degrees=(1,2,6,6))
            sage: type(A)
            <class 'sage.algebras.commutative_dga.GCAlgebra_with_category'>
    """
    @staticmethod
    def __classcall_private__(cls, base, names=None, degrees=None,
                              max_degree=None, category=None, **kwargs):
        r"""
        Normalize the input for the :meth:`__init__` method and the
        unique representation.

        INPUT:

        - ``base`` -- the base ring of the algebra
        - ``max_degree`` -- the maximal degree of the algebra
        - ``names`` -- the names of the variables; by default, set to ``x1``,
          ``x2``, etc
        - ``degrees`` -- the degrees of the generators; by default, set to 1

        TESTS::

            sage: A1 = GradedCommutativeAlgebra(GF(2), 'x,y', (3, 6), max_degree=12)
            sage: A2 = GradedCommutativeAlgebra(GF(2), ['x', 'y'], [3, 6], max_degree=12)
            sage: A1 is A2
            True
        """
        if max_degree is None:
            raise TypeError("max_degree must be specified")
        if names is None:
            if degrees is None:
                raise ValueError("you must specify names or degrees")
            n = len(degrees)
            names = tuple(f'x{i}' for i in range(n))
        elif isinstance(names, str):
            names = tuple(names.split(','))
            n = len(names)
        else:
            n = len(names)
            names = tuple(names)
        if degrees is None:
            degrees = tuple([1 for _ in range(n)])
        else:
            degrees = tuple(degrees)

        return super().__classcall__(cls, base=base, names=names,
                                     degrees=degrees, max_degree=max_degree,
                                     category=category, **kwargs)

    def __init__(self, base, names, degrees, max_degree,
                 category=None, **kwargs):
        r"""
        Construct a commutative graded algebra with finite degree.

        TESTS::

            sage: # needs sage.combinat
            sage: A.<x,y,z,t> = GradedCommutativeAlgebra(QQ, max_degree=6)
            sage: TestSuite(A).run()
            sage: A = GradedCommutativeAlgebra(QQ, ('x','y','z'), [2,3,4], max_degree=8)
            sage: TestSuite(A).run()
            sage: A = GradedCommutativeAlgebra(QQ, ('x','y','z','t'), [1,2,3,4], max_degree=10)
            sage: TestSuite(A).run()
        """
        from sage.arith.misc import gcd

        if max_degree not in ZZ:
            raise TypeError('max_degree must be an integer')
        if max_degree < max(degrees):
            raise ValueError(f'max_degree must not deceed {max(degrees)}')
        self._names = names
        self.__ngens = len(self._names)
        self._degrees = degrees
        self._max_deg = max_degree
        self._weighted_vectors = WeightedIntegerVectors(degrees)
        self._mul_symbol = kwargs.pop('mul_symbol', '*')
        self._mul_latex_symbol = kwargs.pop('mul_latex_symbol', '')
        step = gcd(degrees)
        universe = DisjointUnionEnumeratedSets(self._weighted_vectors.subset(k)
                                               for k in range(0, max_degree, step))
        base_cat = Algebras(base).WithBasis().Super().Supercommutative().FiniteDimensional()
        category = base_cat.or_subcategory(category, join=True)
        indices = ConditionSet(universe, self._valid_index)
        sorting_key = self._weighted_vectors.grading
        CombinatorialFreeModule.__init__(self, base, indices,
                                         sorting_key=sorting_key,
                                         category=category)

    def _valid_index(self, w) -> bool:
        r"""
        Return whether ``w`` is a valid index; no multiple powers in odd
        degrees.

        TESTS::

            sage: A.<x,y,z> = GradedCommutativeAlgebra(QQ, degrees=(1,2,3), max_degree=8)
            sage: w1 = A._weighted_vectors([1,2,1])
            sage: w2 = A._weighted_vectors([1,2,2])
            sage: A._valid_index(w1)
            True
            sage: A._valid_index(w2)
            False
        """
        return not any(i > 1 for i, d in zip(w, self._degrees) if d % 2)

    def _repr_(self) -> str:
        """
        Return the string representation of ``self``.

        TESTS::

            sage: A.<x,y,z> = GradedCommutativeAlgebra(QQ, degrees=(1,2,3), max_degree=8)
            sage: A._repr_()
            "Graded commutative algebra with generators ('x', 'y', 'z') in degrees (1, 2, 3) with maximal degree 8"
            sage: A  # indirect doctest
            Graded commutative algebra with generators ('x', 'y', 'z') in degrees (1, 2, 3) with maximal degree 8
        """
        desc = f'Graded commutative algebra with generators {self._names} in '
        desc += f'degrees {self._degrees} with maximal degree {self._max_deg}'
        return desc

    def ngens(self):
        r"""
        Return the number of generators of ``self``.

        EXAMPLES::

            sage: A.<x,y,z> = GradedCommutativeAlgebra(QQ, degrees=(4,8,2), max_degree=10)
            sage: A.ngens()
            3
        """
        return self.__ngens

    @cached_method
    def product_on_basis(self, w1, w2):
        r"""
        Return the product of two indices within the algebra.

        EXAMPLES::

            sage: A.<x,y,z> = GradedCommutativeAlgebra(QQ, degrees=(4,8,2), max_degree=10)
            sage: z*x
            x*z
            sage: x^3
            0
            sage: 5*z + 4*z*x
            5*z + 4*x*z

        ::

            sage: A.<x,y,z> = GradedCommutativeAlgebra(QQ, degrees=(1,2,3), max_degree=5)
            sage: 2*x*y
            2*x*y
            sage: x^2
            0
            sage: x*z
            x*z
            sage: z*x
            -x*z
            sage: x*y*z
            0

        TESTS::

            sage: A.<x,y,z> = GradedCommutativeAlgebra(QQ, degrees=(4,8,2), max_degree=10)
            sage: weighted_vectors = A._weighted_vectors
            sage: w1 = A._weighted_vectors([1,0,1])
            sage: w2 = A._weighted_vectors([0,0,0])
            sage: A.product_on_basis(w1, w2)
            x*z

        ::

            sage: A.<x,y,z> = GradedCommutativeAlgebra(QQ, degrees=(1,2,3), max_degree=5)
            sage: weighted_vectors = A._weighted_vectors
            sage: w1 = A._weighted_vectors([1,0,0])
            sage: w2 = A._weighted_vectors([0,0,1])
            sage: A.product_on_basis(w1, w2)
            x*z
            sage: A.product_on_basis(w2, w1)
            -x*z

        ::

            sage: A.<x,y,z> = GradedCommutativeAlgebra(QQ, degrees=(1,2,3), max_degree=10)
            sage: weighted_vectors = A._weighted_vectors
            sage: w1 = A._weighted_vectors([1,1,0])
            sage: w2 = A._weighted_vectors([0,1,1])
            sage: A.product_on_basis(w1, w2)
            x*y^2*z
            sage: A.product_on_basis(w2, w1)
            -x*y^2*z
        """
        grading = self._weighted_vectors.grading
        deg_left = grading(w1)
        deg_right = grading(w2)
        deg_tot = deg_left + deg_right
        if deg_tot > self._max_deg:
            return self.zero()
        w_tot = self._weighted_vectors([sum(w) for w in zip(w1, w2)])
        if not self._valid_index(w_tot):
            return self.zero()
        # determine sign
        n = self.__ngens
        c = 0
        for p, i, d in zip(reversed(range(n)), reversed(w1), reversed(self._degrees)):
            if is_even(d) or i == 0:
                continue
            for q, j, b in zip(range(n), w2, self._degrees):
                if q == p:
                    break
                if j == 0 or is_even(b):
                    continue
                c += 1
        return (-1)**c * self.monomial(w_tot)

    def degree_on_basis(self, i):
        r"""
        Return the degree of a homogeneous element with index `i`.

        EXAMPLES::

            sage: A.<a,b,c> = GradedCommutativeAlgebra(QQ, degrees=(2,4,6), max_degree=7)
            sage: a.degree()
            2
            sage: (2*a*b).degree()
            6
            sage: (a+b).degree()
            Traceback (most recent call last):
            ...
            ValueError: element is not homogeneous

        TESTS::

            sage: A.<a,b,c> = GradedCommutativeAlgebra(QQ, degrees=(2,4,6), max_degree=7)
            sage: weighted_vectors = A._weighted_vectors
            sage: i = A._weighted_vectors([1,1,0])
            sage: A.degree_on_basis(i)
            6
        """
        return self._weighted_vectors.grading(i)

    def _repr_term(self, w) -> str:
        r"""
        Return the string representation of basis with index ``w``.

        TESTS::

            sage: A.<x,y,z> = GradedCommutativeAlgebra(QQ, degrees=(1,2,3), max_degree=8)
            sage: w = A._weighted_vectors([1,2,1])
            sage: A._repr_term(w)
            'x*y^2*z'
            sage: x*y^2*z  # indirect doctest
            x*y^2*z

        ::

            sage: A.<x,y,z> = GradedCommutativeAlgebra(QQ, degrees=(1,2,3), max_degree=8, mul_symbol='⌣')
            sage: w = A._weighted_vectors([1,2,1])
            sage: A._repr_term(w)
            'x⌣y^2⌣z'
            sage: x*y^2*z  # indirect doctest
            x⌣y^2⌣z
        """
        # Trivial case:
        if sum(w) == 0:
            return '1'
        # Non-trivial case:
        terms = []
        for i, wi in enumerate(w):
            if wi == 0:
                continue
            if wi == 1:
                terms.append(self._names[i])
            else:
                terms.append(self._names[i] + f'^{wi}')
        return self._mul_symbol.join(terms)

    def _latex_term(self, w) -> str:
        r"""
        Return the LaTeX representation of basis with index ``w``.

        TESTS::

            sage: A.<x,y,z> = GradedCommutativeAlgebra(QQ, degrees=(1,2,3), max_degree=8)
            sage: w = A._weighted_vectors([1,2,1])
            sage: A._latex_term(w)
            'x y^{2} z'
            sage: latex(x*y^2*z)  # indirect doctest
            x y^{2} z

        ::

            sage: A.<x,y,z> = GradedCommutativeAlgebra(QQ, degrees=(1,2,3), max_degree=8, mul_latex_symbol=r'\smile')
            sage: A._latex_term(w)
            'x\\smile y^{2}\\smile z'
            sage: latex(x*y^2*z)  # indirect doctest
            x\smile y^{2}\smile z
        """
        # Trivial case:
        if sum(w) == 0:
            return '1'
        # Non-trivial case:
        terms = []
        for i, wi in enumerate(w):
            if wi == 0:
                continue
            if wi == 1:
                terms.append(self._names[i])
            else:
                terms.append(self._names[i] + '^{' + str(wi) + '}')
        latex_mul = self._mul_latex_symbol + ' '  # add whitespace
        return latex_mul.join(terms)

    def algebra_generators(self):
        r"""
        Return the generators of ``self`` as a
        :class:`sage.sets.family.TrivialFamily`.

        EXAMPLES::

            sage: A.<x,y,z> = GradedCommutativeAlgebra(QQ, degrees=(4,8,2), max_degree=10)
            sage: A.algebra_generators()
            Family (x, y, z)
        """
        from sage.sets.family import Family
        return Family(self.gens())

    @cached_method
    def one_basis(self):
        r"""
        Return the index of the one element of ``self``.

        EXAMPLES::

            sage: A.<x,y,z> = GradedCommutativeAlgebra(QQ, degrees=(4,8,2), max_degree=10)
            sage: ind = A.one_basis(); ind
            [0, 0, 0]
            sage: A.monomial(ind)
            1
            sage: A.one()  # indirect doctest
            1
        """
        n = len(self._degrees)
        return self._weighted_vectors([0 for _ in range(n)])

    def gens(self) -> tuple:
        r"""
        Return the generators of ``self`` as a tuple.

        EXAMPLES::

            sage: A.<x,y,z> = GradedCommutativeAlgebra(QQ, degrees=(4,8,2), max_degree=10)
            sage: A.gens()
            (x, y, z)
        """
        n = len(self._degrees)
        zero = [0 for _ in range(n)]
        indices = []
        for k in range(n):
            ind = list(zero)
            ind[k] = 1
            indices.append(self._weighted_vectors(ind))
        return tuple([self.monomial(ind) for ind in indices])

    @cached_method
    def gen(self, i):
        r"""
        Return the `i`-th generator of ``self``.

        EXAMPLES::

            sage: A.<x,y,z> = GradedCommutativeAlgebra(QQ, degrees=(4,8,2), max_degree=10)
            sage: A.gen(0)
            x
            sage: A.gen(1)
            y
            sage: A.gen(2)
            z
        """
        return self.gens()[i]

    def maximal_degree(self):
        r"""
        Return the maximal degree of ``self``.

        EXAMPLES::

            sage: A.<x,y,z> = GradedCommutativeAlgebra(QQ, degrees=(1,2,3), max_degree=8)
            sage: A.maximal_degree()
            8
        """
        return self._max_deg

    max_degree = maximal_degree



# sage_setup: distribution = sagemath-modules
# sage.doctest: needs sage.modules sage.rings.finite_rings
r"""
Reed-Muller code

Given integers `m, r` and a finite field `F`,
the corresponding Reed-Muller Code is the set:

.. MATH::

    \{ (f(\alpha_i)\mid \alpha_i \in F^m)  \mid  f \in F[x_1,x_2,\ldots,x_m], \deg f \leq r \}

This file contains the following elements:

    - :class:`QAryReedMullerCode`, the class for Reed-Muller codes over non-binary field of size q and `r<q`
    - :class:`BinaryReedMullerCode`, the class for Reed-Muller codes over binary field and `r<=m`
    - :class:`ReedMullerVectorEncoder`, an encoder with a vectorial message space (for both the two code classes)
    - :class:`ReedMullerPolynomialEncoder`, an encoder with a polynomial message space (for both the code classes)
"""
# ****************************************************************************
#       Copyright (C) 2016 Parthasarathi Panda <parthasarathipanda314@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from operator import mul
from functools import reduce

from sage.matrix.constructor import matrix
from sage.arith.misc import binomial
from sage.coding.linear_code import AbstractLinearCode, LinearCodeSyndromeDecoder
from sage.coding.encoder import Encoder
from sage.combinat.subset import Subsets
from sage.combinat.tuple import Tuples
from sage.categories.finite_fields import FiniteFields
from sage.rings.finite_rings.finite_field_constructor import GF
from sage.rings.integer import Integer
from sage.modules.free_module_element import vector
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.misc.cachefunc import cached_method


def _binomial_sum(n, k):
    r"""
    Return the sum of all binomials `\binom{n}{i}`,
    with `i` ranging from `0` to `k` and including `k`.

    INPUT:

    - ``n``, ``k`` -- integers

    EXAMPLES::

        sage: from sage.coding.reed_muller_code import _binomial_sum
        sage: _binomial_sum(4, 2)
        11
    """
    s = 1
    nCi = 1
    for i in range(k):
        nCi = ((n - i) * nCi) // (i + 1)
        s = nCi + s
    return s


def _multivariate_polynomial_interpolation(evaluation, order, polynomial_ring):
    r"""
    Return `f \in \GF{q}[X_1,...,X_m]` such that `f(\mathbf a) = v[i(\mathbf a)]`
    for all `\mathbf a \in \GF{q^m}`, where `v \in \GF{q}^{q^m}` is a given
    vector of evaluations, and `i(a)` is a specific ordering of `\GF{q^m}` (see below for details)

    The ordering `i(a)` is the one used by Sage when listing the elements
    of a Finite Field with a call to the method ``list``.

    In case the polynomial `f` does not exist, this method returns an arbitrary polynomial.

    INPUT:

    - ``evaluation`` -- a vector or a list of evaluation of the polynomial at all the points

    - ``num_of_var`` -- the number of variables used in the polynomial to interpolate

    - ``order`` -- the degree of the polynomial to interpolate

    - ``polynomial_ring`` -- the Polynomial Ring the polynomial in question is from

    EXAMPLES::

        sage: from sage.coding.reed_muller_code import _multivariate_polynomial_interpolation
        sage: F = GF(3)
        sage: R.<x,y> = F[]
        sage: v = vector(F, [1, 2, 0, 0, 2, 1, 1, 1, 1])
        sage: _multivariate_polynomial_interpolation(v, 2, R)
        x*y + y^2 + x + y + 1

    If there does not exist
    """
    def _interpolate(evaluation, num_of_var, order):
        if num_of_var == 0 or order == 0:
            return evaluation[0]
        base_field = polynomial_ring.base_ring()
        q = base_field.cardinality()
        n_by_q = q**(num_of_var - 1)
        d = min(order + 1, q)
        multipoint_evaluation_list = []
        uni_poly_ring = PolynomialRing(base_field, 'x')
        base_field_zero = base_field.zero()
        for k in range(n_by_q):
            iterator = iter(base_field)
            points = []
            for i in range(d):
                xcoordinate = next(iterator)
                points.append((xcoordinate, evaluation[k + i * n_by_q]))
            polyVector = uni_poly_ring.lagrange_polynomial(
                points).coefficients(sparse=False)
            if len(polyVector) < d:
                # adding zeros to represent a (d-1) degree polynomial
                polyVector += [base_field_zero] * (d - len(polyVector))
            multipoint_evaluation_list.append(polyVector)
        poly = polynomial_ring.zero()
        z = 1
        x = polynomial_ring.gen(num_of_var - 1)
        for k in range(d):  # computing the polynomial
            poly = poly + z * _interpolate([multipoint_evaluation_list[i][k]
                                            for i in range(n_by_q)], num_of_var - 1, order - k)
            z *= x
        return poly
    return _interpolate(evaluation, polynomial_ring.ngens(), order)


def ReedMullerCode(base_field, order, num_of_var):
    r"""
    Return a Reed-Muller code.

    A Reed-Muller Code of order `r` and number of variables `m` over a finite field `F` is the set:

    .. MATH::

        \{ (f(\alpha_i)\mid \alpha_i \in F^m)  \mid  f \in F[x_1,x_2,\ldots,x_m], \deg f \leq r \}

    INPUT:

    - ``base_field`` -- the finite field `F` over which the code is built

    - ``order`` -- the order of the Reed-Muller Code, which is the maximum
      degree of the polynomial to be used in the code

    - ``num_of_var`` -- the number of variables used in polynomial

    .. WARNING::

        For now, this implementation only supports Reed-Muller codes whose order is less than q.
        Binary Reed-Muller codes must have their order less than or
        equal to their number of variables.

    EXAMPLES:

    We build a Reed-Muller code::

        sage: F = GF(3)
        sage: C = codes.ReedMullerCode(F, 2, 2)
        sage: C
        Reed-Muller Code of order 2 and 2 variables over Finite Field of size 3

    We ask for its parameters::

        sage: C.length()
        9
        sage: C.dimension()
        6
        sage: C.minimum_distance()
        3

    If one provides a finite field of size 2, a Binary Reed-Muller code is built::

        sage: F = GF(2)
        sage: C = codes.ReedMullerCode(F, 2, 2)
        sage: C
        Binary Reed-Muller Code of order 2 and number of variables 2
    """
    if base_field not in FiniteFields():
        raise ValueError("The parameter `base_field` must be a finite field")
    q = base_field.cardinality()
    if q == 2:
        return BinaryReedMullerCode(order, num_of_var)
    else:
        return QAryReedMullerCode(base_field, order, num_of_var)


class QAryReedMullerCode(AbstractLinearCode):
    r"""
    Representation of a `q`-ary Reed-Muller code.

    For details on the definition of Reed-Muller codes, refer to
    :meth:`ReedMullerCode`.

    .. NOTE::

        It is better to use the aforementioned method rather than calling this
        class directly, as :meth:`ReedMullerCode` creates either a binary or a
        `q`-ary Reed-Muller code according to the arguments it receives.

    INPUT:

    - ``base_field`` -- a finite field, which is the base field of the code

    - ``order`` -- the order of the Reed-Muller Code, i.e., the maximum degree
      of the polynomial to be used in the code

    - ``num_of_var`` -- the number of variables used in polynomial

    .. WARNING::

        For now, this implementation only supports Reed-Muller codes whose order
        is less than q.

    EXAMPLES::

        sage: from sage.coding.reed_muller_code import QAryReedMullerCode
        sage: F = GF(3)
        sage: C = QAryReedMullerCode(F, 2, 2)
        sage: C
        Reed-Muller Code of order 2 and 2 variables over Finite Field of size 3
    """

    _registered_encoders = {}
    _registered_decoders = {}

    def __init__(self, base_field, order, num_of_var):
        r"""
        TESTS:

        Note that the order given cannot be greater than (q-1). An error is raised if that happens::

            sage: from sage.coding.reed_muller_code import QAryReedMullerCode
            sage: C = QAryReedMullerCode(GF(3), 4, 4)
            Traceback (most recent call last):
            ...
            ValueError: The order must be less than 3

        The order and the number of variable must be integers::

            sage: C = QAryReedMullerCode(GF(3),1.1,4)
            Traceback (most recent call last):
            ...
            ValueError: The order of the code must be an integer

        The base_field parameter must be a finite field::

            sage: C = QAryReedMullerCode(QQ,1,4)
            Traceback (most recent call last):
            ...
            ValueError: the input `base_field` must be a FiniteField
        """
        # input sanitization
        if base_field not in FiniteFields():
            raise ValueError("the input `base_field` must be a FiniteField")
        if not isinstance(order, (Integer, int)):
            raise ValueError("The order of the code must be an integer")
        if not isinstance(num_of_var, (Integer, int)):
            raise ValueError("The number of variables must be an integer")
        q = base_field.cardinality()
        if order >= q:
            raise ValueError("The order must be less than %s" % q)

        super().__init__(base_field, q**num_of_var,
                         "EvaluationVector", "Syndrome")
        self._order = order
        self._num_of_var = num_of_var
        self._dimension = binomial(num_of_var + order, order)

    def order(self):
        r"""
        Return the order of ``self``.

        Order is the maximum degree of the polynomial used in the Reed-Muller code.

        EXAMPLES::

            sage: from sage.coding.reed_muller_code import QAryReedMullerCode
            sage: F = GF(59)
            sage: C = QAryReedMullerCode(F, 2, 4)
            sage: C.order()
            2
        """
        return self._order

    def number_of_variables(self):
        r"""
        Return the number of variables of the polynomial ring used in ``self``.

        EXAMPLES::

            sage: from sage.coding.reed_muller_code import QAryReedMullerCode
            sage: F = GF(59)
            sage: C = QAryReedMullerCode(F, 2, 4)
            sage: C.number_of_variables()
            4
        """
        return self._num_of_var

    def minimum_distance(self):
        r"""
        Return the minimum distance between two words in ``self``.

        The minimum distance of a `q`-ary Reed-Muller code with order `d` and
        number of variables `m` is `(q-d)q^{m-1}`

        EXAMPLES::

            sage: from sage.coding.reed_muller_code import QAryReedMullerCode
            sage: F = GF(5)
            sage: C = QAryReedMullerCode(F, 2, 4)
            sage: C.minimum_distance()
            375
        """
        d = self.order()
        q = self.base_field().cardinality()
        n = self.length()
        return ((q - d) * n) // q

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: from sage.coding.reed_muller_code import QAryReedMullerCode
            sage: F = GF(59)
            sage: C = QAryReedMullerCode(F, 2, 4)
            sage: C
            Reed-Muller Code of order 2 and 4 variables over Finite Field of size 59
        """
        return "Reed-Muller Code of order %s and %s variables over %s" % (
            self.order(), self.number_of_variables(), self.base_field())

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: from sage.coding.reed_muller_code import QAryReedMullerCode
            sage: F = GF(59)
            sage: C = QAryReedMullerCode(F, 2, 4)
            sage: latex(C)
            \textnormal{Reed-Muller Code of order} 2 \textnormal{and }4 \textnormal{variables over} \Bold{F}_{59}
        """
        return "\\textnormal{Reed-Muller Code of order} %s \\textnormal{and }%s \\textnormal{variables over} %s"\
            % (self.order(), self.number_of_variables(), self.base_field()._latex_())

    def __eq__(self, other):
        r"""
        Test equality between Reed-Muller Code objects.

        EXAMPLES::

            sage: from sage.coding.reed_muller_code import QAryReedMullerCode
            sage: F = GF(59)
            sage: C1 = QAryReedMullerCode(F, 2, 4)
            sage: C2 = QAryReedMullerCode(GF(59), 2, 4)
            sage: C1.__eq__(C2)
            True
        """
        # I am not comparing the base field directly because of possible change
        # in variables
        return isinstance(other, QAryReedMullerCode) \
            and self.base_field() == other.base_field() \
            and self.order() == other.order() \
            and self.number_of_variables() == other.number_of_variables()


class BinaryReedMullerCode(AbstractLinearCode):
    r"""
    Representation of a binary Reed-Muller code.

    For details on the definition of Reed-Muller codes, refer to
    :meth:`ReedMullerCode`.

    .. NOTE::

        It is better to use the aforementioned method rather than calling this
        class directly, as :meth:`ReedMullerCode` creates either a binary or a
        `q`-ary Reed-Muller code according to the arguments it receives.


    INPUT:

    - ``order`` -- the order of the Reed-Muller Code, i.e., the maximum degree
      of the polynomial to be used in the code

    - ``num_of_var`` -- the number of variables used in the polynomial

    EXAMPLES:

    A binary Reed-Muller code can be constructed by simply giving the order of
    the code and the number of variables::

        sage: C = codes.BinaryReedMullerCode(2, 4)
        sage: C
        Binary Reed-Muller Code of order 2 and number of variables 4

    Very large Reed-Muller codes can be constructed without building
    the generator matrix or elements of the code (fixes :issue:`33229`,
    see also :issue:`39110`)::

        sage: C = codes.BinaryReedMullerCode(16, 32)
        sage: C
        Binary Reed-Muller Code of order 16 and number of variables 32
        sage: C.dimension(), C.length()
        (2448023843, 4294967296)
    """

    _registered_encoders = {}
    _registered_decoders = {}

    def __init__(self, order, num_of_var):
        r"""
        TESTS:

        If the order given is greater than the number of variables an error is raised::

            sage: C = codes.BinaryReedMullerCode(5, 4)
            Traceback (most recent call last):
            ...
            ValueError: The order must be less than or equal to 4

        The order and the number of variable must be integers::

            sage: C = codes.BinaryReedMullerCode(1.1,4)
            Traceback (most recent call last):
            ...
            ValueError: The order of the code must be an integer
        """
        # input sanitization
        if not isinstance(order, (Integer, int)):
            raise ValueError("The order of the code must be an integer")
        if not isinstance(num_of_var, (Integer, int)):
            raise ValueError("The number of variables must be an integer")
        if (num_of_var < order):
            raise ValueError(
                "The order must be less than or equal to %s" %
                num_of_var)

        super().__init__(GF(2), 2**num_of_var, "EvaluationVector", "Syndrome")
        self._order = order
        self._num_of_var = num_of_var
        self._dimension = _binomial_sum(num_of_var, order)

    def order(self):
        r"""
        Return the order of ``self``.

        Order is the maximum degree of the polynomial used in the Reed-Muller code.

        EXAMPLES::

            sage: C = codes.BinaryReedMullerCode(2, 4)
            sage: C.order()
            2
        """
        return self._order

    def number_of_variables(self):
        r"""
        Return the number of variables of the polynomial ring used in ``self``.

        EXAMPLES::

            sage: C = codes.BinaryReedMullerCode(2, 4)
            sage: C.number_of_variables()
            4
        """
        return self._num_of_var

    def minimum_distance(self):
        r"""
        Return the minimum distance of ``self``.

        The minimum distance of a binary Reed-Muller code of order `d` and
        number of variables `m` is `q^{m-d}`

        EXAMPLES::

            sage: C = codes.BinaryReedMullerCode(2, 4)
            sage: C.minimum_distance()
            4
        """
        return 2**(self.number_of_variables() - self.order())

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: C = codes.BinaryReedMullerCode(2, 4)
            sage: C
            Binary Reed-Muller Code of order 2 and number of variables 4
        """
        return "Binary Reed-Muller Code of order %s and number of variables %s" % (
            self.order(), self.number_of_variables())

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: C = codes.BinaryReedMullerCode(2, 4)
            sage: latex(C)
            \textnormal{Binary Reed-Muller Code of order} 2 \textnormal{and number of variables} 4
        """
        return "\\textnormal{Binary Reed-Muller Code of order} %s \\textnormal{and number of variables} %s" % (
            self.order(), self.number_of_variables())

    def __eq__(self, other):
        r"""
        Test equality between Reed-Muller Code objects.

        EXAMPLES::

            sage: C1 = codes.BinaryReedMullerCode(2, 4)
            sage: C2 = codes.BinaryReedMullerCode(2, 4)
            sage: C1.__eq__(C2)
            True
        """
        return isinstance(other, BinaryReedMullerCode) \
            and self.order() == other.order() \
            and self.number_of_variables() == other.number_of_variables()


class ReedMullerVectorEncoder(Encoder):
    r"""
    Encoder for Reed-Muller codes which encodes vectors into codewords.

    Consider a Reed-Muller code of order `r`, number of variables `m`, length `n`,
    dimension `k` over some finite field `F`.
    Let those variables be `(x_1, x_2, \dots, x_m)`.
    We order the monomials by lowest power on lowest index variables. If we have
    three monomials `x_1 x_2`, `x_1 x_2^2` and `x_1^2 x_2`, the ordering is:
    `x_1 x_2 < x_1 x_2^2 < x_1^2 x_2`

    Let now `(v_1,v_2,\ldots,v_k)` be a vector of `F`, which corresponds to the polynomial
    `f = \Sigma^{k}_{i=1} v_i x_i`.

    Let `(\beta_1, \beta_2, \ldots, \beta_q)` be the elements of `F` ordered as they are
    returned by Sage when calling ``F.list()``.

    The aforementioned polynomial `f` is encoded as:

    `(f(\alpha_{11},\alpha_{12},\ldots,\alpha_{1m}),f(\alpha_{21},\alpha_{22},\ldots,
    \alpha_{2m}),\ldots,f(\alpha_{q^m1},\alpha_{q^m2},\ldots,\alpha_{q^mm}))`

    with `\alpha_{ij}=\beta_{i \bmod{q^j}}` for all `i`, `j)`.

    INPUT:

    - ``code`` -- the associated code of this encoder

    EXAMPLES::

        sage: C1 = codes.ReedMullerCode(GF(2), 2, 4)
        sage: E1 = codes.encoders.ReedMullerVectorEncoder(C1)
        sage: E1
        Evaluation vector-style encoder for
         Binary Reed-Muller Code of order 2 and number of variables 4
        sage: C2 = codes.ReedMullerCode(GF(3), 2, 2)
        sage: E2 = codes.encoders.ReedMullerVectorEncoder(C2)
        sage: E2
        Evaluation vector-style encoder for
         Reed-Muller Code of order 2 and 2 variables over Finite Field of size 3

    Actually, we can construct the encoder from ``C`` directly::

        sage: C=codes.ReedMullerCode(GF(2), 2, 4)
        sage: E = C.encoder("EvaluationVector")
        sage: E
        Evaluation vector-style encoder for
         Binary Reed-Muller Code of order 2 and number of variables 4
    """

    def __init__(self, code):
        r"""
        TESTS:

        If ``code`` is not a Reed-Muller code, an error is raised::

            sage: C  = codes.random_linear_code(GF(11), 10, 4)
            sage: codes.encoders.ReedMullerVectorEncoder(C)
            Traceback (most recent call last):
            ...
            ValueError: the code has to be a Reed-Muller code
        """
        if not isinstance(code, (QAryReedMullerCode, BinaryReedMullerCode)):
            raise ValueError("the code has to be a Reed-Muller code")
        super().__init__(code)

    def _repr_(self) -> str:
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: F = GF(11)
            sage: C = codes.ReedMullerCode(F, 2, 4)
            sage: E=codes.encoders.ReedMullerVectorEncoder(C)
            sage: E
            Evaluation vector-style encoder for Reed-Muller Code of order 2 and 4 variables over Finite Field of size 11
        """
        return "Evaluation vector-style encoder for %s" % self.code()

    def _latex_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: F = GF(11)
            sage: C = codes.ReedMullerCode(F, 2, 4)
            sage: E=codes.encoders.ReedMullerVectorEncoder(C)
            sage: latex(E)
            \textnormal{Evaluation vector-style encoder for }\textnormal{Reed-Muller Code of order} 2 \textnormal{and }4 \textnormal{variables over} \Bold{F}_{11}
        """
        return "\\textnormal{Evaluation vector-style encoder for }%s" % self.code()._latex_()

    def __eq__(self, other):
        r"""
        Test equality between ReedMullerVectorEncoder objects.

        EXAMPLES::

            sage: F = GF(11)
            sage: C = codes.ReedMullerCode(F, 2, 4)
            sage: D1 = codes.encoders.ReedMullerVectorEncoder(C)
            sage: D2 = codes.encoders.ReedMullerVectorEncoder(C)
            sage: D1.__eq__(D2)
            True
            sage: D1 is D2
            False
        """
        return (isinstance(other, ReedMullerVectorEncoder)
                ) and self.code() == other.code()

    @cached_method
    def generator_matrix(self):
        r"""
        Return a generator matrix of ``self``.

        EXAMPLES::

            sage: F = GF(3)
            sage: C = codes.ReedMullerCode(F, 2, 2)
            sage: E = codes.encoders.ReedMullerVectorEncoder(C)
            sage: E.generator_matrix()
            [1 1 1 1 1 1 1 1 1]
            [0 1 2 0 1 2 0 1 2]
            [0 0 0 1 1 1 2 2 2]
            [0 1 1 0 1 1 0 1 1]
            [0 0 0 0 1 2 0 2 1]
            [0 0 0 1 1 1 1 1 1]
        """
        C = self.code()
        base_field = C.base_field()
        order = C.order()
        num_of_var = C.number_of_variables()
        q = base_field.cardinality()
        points = base_field**num_of_var
        matrix_list = []
        max_individual_degree = min(order, (q - 1))
        for degree in range(order + 1):
            exponents = Subsets(list(range(num_of_var)) * max_individual_degree,
                                degree, submultiset=True)
            matrix_list += [[reduce(mul, [x[i] for i in exponent], 1)
                             for x in points] for exponent in exponents]
        M = matrix(base_field, matrix_list)
        M.set_immutable()
        return M

    def points(self):
        r"""
        Return the points of `F^m`, where `F` is the base field and `m` is the
        number of variables, in order of which polynomials are evaluated on.

        EXAMPLES::

            sage: F = GF(3)
            sage: Fx.<x0,x1> = F[]
            sage: C = codes.ReedMullerCode(F, 2, 2)
            sage: E = C.encoder("EvaluationVector")
            sage: E.points()
            [(0, 0), (1, 0), (2, 0), (0, 1), (1, 1), (2, 1), (0, 2), (1, 2), (2, 2)]
        """
        code = self.code()
        return ((code.base_field())**code.number_of_variables()).list()


class ReedMullerPolynomialEncoder(Encoder):
    r"""
    Encoder for Reed-Muller codes which encodes appropriate multivariate polynomials into codewords.

    Consider a Reed-Muller code of order `r`, number of variables `m`, length `n`,
    dimension `k` over some finite field `F`.
    Let those variables be `(x_1, x_2, \dots, x_m)`.
    We order the monomials by lowest power on lowest index variables. If we have three monomials
    `x_1 x_2`, `x_1 x_2^2` and `x_1^2 x_2`, the ordering is:
    `x_1 x_2 < x_1 x_2^2 < x_1^2 x_2`

    Let now `f` be a polynomial of the multivariate polynomial ring `F[x_1, \dots, x_m]`.

    Let `(\beta_1, \beta_2, \ldots, \beta_q)` be the elements of `F` ordered as they are
    returned by Sage when calling ``F.list()``.

    The aforementioned polynomial `f` is encoded as:

    `(f(\alpha_{11},\alpha_{12},\ldots,\alpha_{1m}),f(\alpha_{21},\alpha_{22},\ldots,
    \alpha_{2m}),\ldots,f(\alpha_{q^m1},\alpha_{q^m2},\ldots,\alpha_{q^mm}))`

    with `\alpha_{ij}=\beta_{i \bmod{q^j}}` for all `i`, `j`.

    INPUT:

    - ``code`` -- the associated code of this encoder

    - ``polynomial_ring`` -- (default: ``None``) the polynomial ring from which
      the message is chosen;  if this is set to ``None``, a polynomial ring in
      `x` will be built from the code parameters

    EXAMPLES::

        sage: C1 = codes.ReedMullerCode(GF(2), 2, 4)
        sage: E1 = codes.encoders.ReedMullerPolynomialEncoder(C1)
        sage: E1
        Evaluation polynomial-style encoder for
         Binary Reed-Muller Code of order 2 and number of variables 4
        sage: C2 = codes.ReedMullerCode(GF(3), 2, 2)
        sage: E2 = codes.encoders.ReedMullerPolynomialEncoder(C2)
        sage: E2
        Evaluation polynomial-style encoder for
         Reed-Muller Code of order 2 and 2 variables over Finite Field of size 3

    We can also pass a predefined polynomial ring::

        sage: R = PolynomialRing(GF(3), 2, 'y')
        sage: C = codes.ReedMullerCode(GF(3), 2, 2)
        sage: E = codes.encoders.ReedMullerPolynomialEncoder(C, R)
        sage: E
        Evaluation polynomial-style encoder for
         Reed-Muller Code of order 2 and 2 variables over Finite Field of size 3

    Actually, we can construct the encoder from ``C`` directly::

        sage: E = C1.encoder("EvaluationPolynomial")
        sage: E
        Evaluation polynomial-style encoder for
         Binary Reed-Muller Code of order 2 and number of variables 4
    """

    def __init__(self, code, polynomial_ring=None):
        r"""
        TESTS:

        If ``code`` is not a Reed-Muller code, an error is raised::

            sage: C  = codes.random_linear_code(GF(11), 10, 4)
            sage: codes.encoders.ReedMullerPolynomialEncoder(C)
            Traceback (most recent call last):
            ...
            ValueError: the code has to be a Reed-Muller code

        If the polynomial ring passed is not according to the requirement (over a different field or different number of variables) then an error is raised::

            sage: F=GF(59)
            sage: R.<x,y,z,w>=F[]
            sage: C=codes.ReedMullerCode(F, 2, 3)
            sage: E=codes.encoders.ReedMullerPolynomialEncoder(C, R)
            Traceback (most recent call last):
            ...
            ValueError: The Polynomial ring should be on Finite Field of size 59 and should have 3 variables
        """
        if not isinstance(code, (QAryReedMullerCode, BinaryReedMullerCode)):
            raise ValueError("the code has to be a Reed-Muller code")
        super().__init__(code)
        if polynomial_ring is None:
            self._polynomial_ring = PolynomialRing(code.base_field(),
                    code.number_of_variables(), 'x')
        else:
            if (polynomial_ring.base_ring() == code.base_field()) and (
                    len(polynomial_ring.variable_names()) == code.number_of_variables()):
                self._polynomial_ring = polynomial_ring
            else:
                raise ValueError(
                    "The Polynomial ring should be on %s and should have %s variables" %
                    (code.base_field(), code.number_of_variables()))

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: F = GF(59)
            sage: C = codes.ReedMullerCode(F, 2, 4)
            sage: E=codes.encoders.ReedMullerPolynomialEncoder(C)
            sage: E
            Evaluation polynomial-style encoder for Reed-Muller Code of order 2 and 4 variables over Finite Field of size 59
        """
        return "Evaluation polynomial-style encoder for %s" % self.code()

    def _latex_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: F = GF(59)
            sage: C = codes.ReedMullerCode(F, 2, 4)
            sage: E=codes.encoders.ReedMullerPolynomialEncoder(C)
            sage: latex(E)
            \textnormal{Evaluation polynomial-style encoder for }\textnormal{Reed-Muller Code of order} 2 \textnormal{and }4 \textnormal{variables over} \Bold{F}_{59}
        """
        return "\\textnormal{Evaluation polynomial-style encoder for }%s" % self.code()._latex_()

    def __eq__(self, other):
        r"""
        Test equality between ReedMullerVectorEncoder objects.

        EXAMPLES::

            sage: F = GF(11)
            sage: C = codes.ReedMullerCode(F, 2, 4)
            sage: D1 = codes.encoders.ReedMullerPolynomialEncoder(C)
            sage: D2 = codes.encoders.ReedMullerPolynomialEncoder(C)
            sage: D1.__eq__(D2)
            True
            sage: D1 is D2
            False
        """
        return isinstance(other, ReedMullerPolynomialEncoder) \
            and self.code() == other.code()

    def encode(self, p):
        r"""
        Transform the polynomial ``p`` into a codeword of :meth:`code`.

        INPUT:

        - ``p`` -- a polynomial from the message space of ``self`` of degree
          less than ``self.code().order()``

        OUTPUT: a codeword in associated code of ``self``

        EXAMPLES::

            sage: F = GF(3)
            sage: Fx.<x0,x1> = F[]
            sage: C = codes.ReedMullerCode(F, 2, 2)
            sage: E = C.encoder("EvaluationPolynomial")
            sage: p = x0*x1 + x1^2 + x0 + x1 + 1
            sage: c = E.encode(p); c
            (1, 2, 0, 0, 2, 1, 1, 1, 1)
            sage: c in C
            True

        If a polynomial with good monomial degree but wrong monomial
        degree is given, an error is raised::

            sage: p = x0^2*x1
            sage: E.encode(p)
            Traceback (most recent call last):
            ...
            ValueError: The polynomial to encode must have degree at most 2

        If ``p`` is not an element of the proper polynomial ring, an error is raised::

            sage: Qy.<y1,y2> = QQ[]
            sage: p = y1^2 + 1
            sage: E.encode(p)
            Traceback (most recent call last):
            ...
            ValueError: The value to encode must be in
            Multivariate Polynomial Ring in x0, x1 over Finite Field of size 3
        """
        M = self.message_space()
        if p not in M:
            raise ValueError("The value to encode must be in %s" % M)
        C = self.code()
        if p.degree() > C.order():
            raise ValueError("The polynomial to encode must have degree at most %s"
                             % C.order())
        base_fieldTuple = Tuples(C.base_field().list(), C.number_of_variables())
        return vector(C.base_ring(), [p(x) for x in base_fieldTuple])

    def unencode_nocheck(self, c):
        r"""
        Return the message corresponding to the codeword ``c``.

        Use this method with caution: it does not check if ``c``
        belongs to the code, and if this is not the case, the output is
        unspecified. Instead, use :meth:`unencode`.

        INPUT:

        - ``c`` -- a codeword of :meth:`code`

        OUTPUT:

        - A polynomial of degree less than ``self.code().order()``.

        EXAMPLES::

            sage: F = GF(3)
            sage: C = codes.ReedMullerCode(F, 2, 2)
            sage: E = C.encoder("EvaluationPolynomial")
            sage: c = vector(F, (1, 2, 0, 0, 2, 1, 1, 1, 1))
            sage: c in C
            True
            sage: p = E.unencode_nocheck(c); p
            x0*x1 + x1^2 + x0 + x1 + 1
            sage: E.encode(p) == c
            True

        Note that no error is thrown if ``c`` is not a codeword, and that the
        result is undefined::

            sage: c = vector(F, (1, 2, 0, 0, 2, 1, 0, 1, 1))
            sage: c in C
            False
            sage: p = E.unencode_nocheck(c); p
            -x0*x1 - x1^2 + x0 + 1
            sage: E.encode(p) == c
            False
        """
        return _multivariate_polynomial_interpolation(
            c,
            self.code().order(),
            self.polynomial_ring())

    def message_space(self):
        r"""
        Return the message space of ``self``.

        EXAMPLES::

            sage: F = GF(11)
            sage: C = codes.ReedMullerCode(F, 2, 4)
            sage: E = C.encoder("EvaluationPolynomial")
            sage: E.message_space()
            Multivariate Polynomial Ring in x0, x1, x2, x3 over Finite Field of size 11
        """
        return self._polynomial_ring

    def polynomial_ring(self):
        r"""
        Return the polynomial ring associated with ``self``.

        EXAMPLES::

            sage: F = GF(11)
            sage: C = codes.ReedMullerCode(F, 2, 4)
            sage: E = C.encoder("EvaluationPolynomial")
            sage: E.polynomial_ring()
            Multivariate Polynomial Ring in x0, x1, x2, x3 over Finite Field of size 11
        """
        return self._polynomial_ring

    def points(self):
        r"""
        Return the evaluation points in the appropriate order as used by ``self`` when
        encoding a message.

        EXAMPLES::

            sage: F = GF(3)
            sage: Fx.<x0,x1> = F[]
            sage: C = codes.ReedMullerCode(F, 2, 2)
            sage: E = C.encoder("EvaluationPolynomial")
            sage: E.points()
            [(0, 0), (1, 0), (2, 0), (0, 1), (1, 1), (2, 1), (0, 2), (1, 2), (2, 2)]
        """
        code = self.code()
        return ((code.base_field())**code.number_of_variables()).list()


# --------------- registration --------------

QAryReedMullerCode._registered_encoders["EvaluationVector"] = ReedMullerVectorEncoder
QAryReedMullerCode._registered_encoders["EvaluationPolynomial"] = ReedMullerPolynomialEncoder

QAryReedMullerCode._registered_decoders["Syndrome"] = LinearCodeSyndromeDecoder

BinaryReedMullerCode._registered_encoders["EvaluationVector"] = ReedMullerVectorEncoder
BinaryReedMullerCode._registered_encoders["EvaluationPolynomial"] = ReedMullerPolynomialEncoder

BinaryReedMullerCode._registered_decoders["Syndrome"] = LinearCodeSyndromeDecoder
